import numpy as np
import numpy.testing as npt

from hats.pixel_math.healpix_pixel_function import get_pixel_argsort, sort_pixels


def test_get_pixel_argsort(pixel_list_depth_first, pixel_list_breadth_first):
    argsort = get_pixel_argsort(pixel_list_depth_first)
    npt.assert_array_equal(argsort, [6, 7, 8, 1, 2, 0, 3, 4, 5])
    sorted_pixel_list = np.array(pixel_list_depth_first)[argsort]
    npt.assert_array_equal(sorted_pixel_list, pixel_list_breadth_first)


def test_get_pixel_argsort_empty():
    argsort = get_pixel_argsort([])
    npt.assert_array_equal(argsort, [])

    argsort = get_pixel_argsort(None)
    npt.assert_array_equal(argsort, [])

    argsort = get_pixel_argsort(np.array([]))
    npt.assert_array_equal(argsort, [])


def test_sort_pixels(pixel_list_depth_first, pixel_list_breadth_first):
    sort_result = sort_pixels(pixel_list_depth_first)
    npt.assert_array_equal(sort_result, pixel_list_breadth_first)

    sort_result = sort_pixels(pixel_list_depth_first)
    npt.assert_array_equal(sort_result, pixel_list_breadth_first)


def test_sort_pixels_empty():
    sort_result = sort_pixels([])
    npt.assert_array_equal(sort_result, [])

    sort_result = sort_pixels(None)
    npt.assert_array_equal(sort_result, [])

    sort_result = sort_pixels(np.array([]))
    npt.assert_array_equal(sort_result, [])



"""Utility functions for handling parquet metadata files"""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import List

import numpy as np
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.parquet as pq
from upath import UPath

from hats.io import file_io, paths
from hats.io.file_io.file_pointer import get_upath
from hats.pixel_math.healpix_pixel import INVALID_PIXEL, HealpixPixel
from hats.pixel_math.healpix_pixel_function import get_pixel_argsort


def row_group_stat_single_value(row_group, stat_key: str):
    """Convenience method to find the min and max inside a statistics dictionary,
    and raise an error if they're unequal.

    Args:
        row_group: dataset fragment row group
        stat_key (str): column name of interest.
    Returns:
        The value of the specified row group statistic
    """
    if stat_key not in row_group.statistics:
        raise ValueError(f"row group doesn't have expected key {stat_key}")
    stat_dict = row_group.statistics[stat_key]
    min_val = stat_dict["min"]
    max_val = stat_dict["max"]
    if min_val != max_val:
        raise ValueError(f"stat min != max ({min_val} != {max_val})")
    return min_val


def get_healpix_pixel_from_metadata(
    metadata: pq.FileMetaData, norder_column: str = "Norder", npix_column: str = "Npix"
) -> HealpixPixel:
    """Get the healpix pixel according to a parquet file's metadata.

    This is determined by the value of Norder and Npix in the table's data

    Args:
        metadata (pyarrow.parquet.FileMetaData): full metadata for a single file.

    Returns:
        Healpix pixel representing the Norder and Npix from the first row group.
    """
    if metadata.num_row_groups <= 0 or metadata.num_columns <= 0:
        raise ValueError("metadata is for empty table")
    order = -1
    pixel = -1
    first_row_group = metadata.row_group(0)
    for i in range(0, first_row_group.num_columns):
        column = first_row_group.column(i)
        if column.path_in_schema == norder_column:
            if column.statistics.min != column.statistics.max:
                raise ValueError(
                    f"{norder_column} stat min != max ({column.statistics.min} != {column.statistics.max})"
                )
            order = column.statistics.min
        elif column.path_in_schema == npix_column:
            if column.statistics.min != column.statistics.max:
                raise ValueError(
                    f"{npix_column} stat min != max ({column.statistics.min} != {column.statistics.max})"
                )
            pixel = column.statistics.min

    if order == -1 or pixel == -1:
        raise ValueError(f"Metadata missing Norder ({norder_column}) or Npix ({npix_column}) column")
    return HealpixPixel(order, pixel)


def write_parquet_metadata(
    catalog_path: str | Path | UPath,
    order_by_healpix=True,
    output_path: str | Path | UPath | None = None,
):
    """Generate parquet metadata, using the already-partitioned parquet files
    for this catalog.

    For more information on the general parquet metadata files, and why we write them, see
    https://arrow.apache.org/docs/python/parquet.html#writing-metadata-and-common-metadata-files

    Args:
        catalog_path (str): base path for the catalog
        order_by_healpix (bool): use False if the dataset is not to be reordered by
            breadth-first healpix pixel (e.g. secondary indexes)
        output_path (str): base path for writing out metadata files
            defaults to `catalog_path` if unspecified

    Returns:
        sum of the number of rows in the dataset.
    """
    ignore_prefixes = [
        "intermediate",
        "_common_metadata",
        "_metadata",
    ]

    catalog_path = get_upath(catalog_path)
    dataset_subdir = catalog_path / "dataset"
    (dataset_path, dataset) = file_io.read_parquet_dataset(
        dataset_subdir,
        ignore_prefixes=ignore_prefixes,
        exclude_invalid_files=True,
    )
    metadata_collector = []
    # Collect the healpix pixels so we can sort before writing.
    healpix_pixels = []
    total_rows = 0

    for single_file in dataset.files:
        relative_path = single_file[len(dataset_path) + 1 :]
        single_metadata = file_io.read_parquet_metadata(dataset_subdir / relative_path)

        # Users must set the file path of each chunk before combining the metadata.
        single_metadata.set_file_path(relative_path)

        if order_by_healpix:
            healpix_pixel = paths.get_healpix_from_path(relative_path)
            if healpix_pixel == INVALID_PIXEL:
                healpix_pixel = get_healpix_pixel_from_metadata(single_metadata)

            healpix_pixels.append(healpix_pixel)
        metadata_collector.append(single_metadata)
        total_rows += single_metadata.num_rows

    ## Write out the two metadata files
    if output_path is None:
        output_path = catalog_path
    if order_by_healpix:
        argsort = get_pixel_argsort(healpix_pixels)
        metadata_collector = np.array(metadata_collector)[argsort]
    catalog_base_dir = get_upath(output_path)
    file_io.make_directory(catalog_base_dir / "dataset", exist_ok=True)
    metadata_file_pointer = paths.get_parquet_metadata_pointer(catalog_base_dir)
    common_metadata_file_pointer = paths.get_common_metadata_pointer(catalog_base_dir)

    file_io.write_parquet_metadata(
        dataset.schema,
        metadata_file_pointer,
        metadata_collector=metadata_collector,
        write_statistics=True,
    )
    file_io.write_parquet_metadata(dataset.schema, common_metadata_file_pointer)
    return total_rows


def write_parquet_metadata_for_batches(batches: List[List[pa.RecordBatch]], output_path: str = None):
    """Write parquet metadata files for some pyarrow table batches.
    This writes the batches to a temporary parquet dataset using local storage, and
    generates the metadata for the partitioned catalog parquet files.

    Args:
        batches (List[List[pa.RecordBatch]]): create one row group per RecordBatch, grouped
            into tables by the inner list.
        output_path (str): base path for writing out metadata files
            defaults to `catalog_path` if unspecified

    Returns:
        sum of the number of rows in the dataset.
    """

    with tempfile.TemporaryDirectory() as temp_pq_file:
        temp_dataset_dir = get_upath(temp_pq_file) / "dataset"
        temp_dataset_dir.mkdir()
        for batch_list in batches:
            temp_info_table = pa.Table.from_batches(batch_list)
            pq.write_to_dataset(temp_info_table, temp_dataset_dir)
        return write_parquet_metadata(temp_pq_file, output_path=output_path)


def read_row_group_fragments(metadata_file: str):
    """Generator for metadata fragment row groups in a parquet metadata file.

    Args:
        metadata_file (str): path to `_metadata` file.
    """
    metadata_file = get_upath(metadata_file)
    if not file_io.is_regular_file(metadata_file):
        metadata_file = paths.get_parquet_metadata_pointer(metadata_file)

    dataset = pds.parquet_dataset(metadata_file, filesystem=metadata_file.fs)

    for frag in dataset.get_fragments():
        yield from frag.row_groups



"""Two sample benchmarks to compute runtime and memory usage.

For more information on writing benchmarks:
https://asv.readthedocs.io/en/stable/writing_benchmarks.html."""

import os

import numpy as np

import hats.pixel_math as hist
import hats.pixel_math.healpix_shim as hp
from hats.catalog import Catalog, PartitionInfo, TableProperties
from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.io.paths import pixel_catalog_files
from hats.pixel_math import HealpixPixel
from hats.pixel_tree import PixelAlignment, align_trees
from hats.pixel_tree.pixel_tree import PixelTree


def time_test_alignment_even_sky():
    """Create alignment from an even distribution at order 7"""
    initial_histogram = np.full(hp.order2npix(7), 40)
    result = hist.generate_alignment(initial_histogram, highest_order=7, threshold=1_000)
    # everything maps to order 5, given the density
    for mapping in result:
        assert mapping[0] == 5


def time_test_cone_filter_multiple_order():
    """Create a catalog cone filter where we have multiple orders in the catalog"""
    catalog_info = TableProperties(
        **{
            "catalog_name": "test_name",
            "catalog_type": "object",
            "total_rows": 10,
            "ra_column": "ra",
            "dec_column": "dec",
        }
    )
    pixels = [HealpixPixel(6, 30), HealpixPixel(7, 124), HealpixPixel(7, 5000)]
    catalog = Catalog(catalog_info, pixels)
    filtered_catalog = catalog.filter_by_cone(47.1, 6, 30 * 3600)
    assert filtered_catalog.get_healpix_pixels() == [HealpixPixel(6, 30), HealpixPixel(7, 124)]


class Suite:
    def __init__(self) -> None:
        """Just initialize things"""
        self.pixel_list = None
        self.pixel_tree_1 = None
        self.pixel_tree_2 = None

    def setup(self):
        self.pixel_list = [HealpixPixel(8, pixel) for pixel in np.arange(100000)]
        self.pixel_tree_1 = PixelTree.from_healpix(self.pixel_list)
        self.pixel_tree_2 = PixelTree.from_healpix(
            [HealpixPixel(9, pixel) for pixel in np.arange(0, 400000, 4)]
        )

    def time_pixel_tree_creation(self):
        PixelTree.from_healpix(self.pixel_list)

    def time_inner_pixel_alignment(self):
        align_trees(self.pixel_tree_1, self.pixel_tree_2)

    def time_outer_pixel_alignment(self):
        align_trees(self.pixel_tree_1, self.pixel_tree_2, alignment_type="outer")

    def time_paths_creation(self):
        pixel_catalog_files("foo/", self.pixel_list)


class MetadataSuite:
    """Suite that generates catalog files and benchmarks the operations on them."""

    def setup_cache(self):
        root_dir = os.getcwd()

        ## Create partition info for catalog a (only at order 7)
        pixel_list_a = [HealpixPixel(7, pixel) for pixel in np.arange(100_000)]
        catalog_path_a = os.path.join(root_dir, "catalog_a")
        os.makedirs(catalog_path_a, exist_ok=True)
        partition_info = PartitionInfo.from_healpix(pixel_list_a)
        partition_info.write_to_file(os.path.join(catalog_path_a, "partition_info.csv"))

        ## Create partition info for catalog a (only at order 6)
        pixel_list_b = [HealpixPixel(6, pixel) for pixel in np.arange(25_000)]
        catalog_path_b = os.path.join(root_dir, "catalog_b")
        os.makedirs(catalog_path_b, exist_ok=True)
        partition_info = PartitionInfo.from_healpix(pixel_list_b)
        partition_info.write_to_file(os.path.join(catalog_path_b, "partition_info.csv"))

        ## Fake an association table between the two
        tree_a = PixelTree.from_healpix(pixel_list_a)
        tree_b = PixelTree.from_healpix(pixel_list_b)
        alignment = align_trees(tree_a, tree_b)
        alignment_df = alignment.pixel_mapping[
            [
                PixelAlignment.PRIMARY_ORDER_COLUMN_NAME,
                PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME,
                PixelAlignment.JOIN_ORDER_COLUMN_NAME,
                PixelAlignment.JOIN_PIXEL_COLUMN_NAME,
            ]
        ]
        alignment_df = alignment_df.rename(
            columns={
                PixelAlignment.PRIMARY_ORDER_COLUMN_NAME: "Norder",
                PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME: "Npix",
            }
        )

        association_catalog_path = os.path.join(root_dir, "assocation_a_b")
        os.makedirs(association_catalog_path, exist_ok=True)
        partition_info = PartitionJoinInfo(alignment_df)
        partition_info.write_to_csv(association_catalog_path)

        return (catalog_path_a, catalog_path_b, association_catalog_path)

    def time_load_partition_info_order7(self, cache):
        PartitionInfo.read_from_dir(cache[0])

    def time_load_partition_info_order6(self, cache):
        PartitionInfo.read_from_dir(cache[1])

    def time_load_partition_join_info(self, cache):
        PartitionInfo.read_from_dir(cache[2])



from __future__ import annotations

import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import Any, Tuple

import numpy as np
import pandas as pd
import pyarrow.dataset as pds
import pyarrow.parquet as pq
import yaml
from pyarrow.dataset import Dataset
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.io.file_io.file_pointer import get_upath


def make_directory(file_pointer: str | Path | UPath, exist_ok: bool = False):
    """Make a directory at a given file pointer

    Will raise an error if a directory already exists, unless `exist_ok` is True in which case
    any existing directories will be left unmodified

    Args:
        file_pointer: location in file system to make directory
        exist_ok: Default False. If false will raise error if directory exists. If true existing
            directories will be ignored and not modified

    Raises:
        OSError
    """
    file_pointer = get_upath(file_pointer)
    file_pointer.mkdir(parents=True, exist_ok=exist_ok)


def _rmdir_recursive(directory):
    for item in directory.iterdir():
        if item.is_dir():
            _rmdir_recursive(item)
        else:
            item.unlink()
    directory.rmdir()


def remove_directory(file_pointer: str | Path | UPath, ignore_errors=False):
    """Remove a directory, and all contents, recursively.

    Args:
        file_pointer: directory in file system to remove
        ignore_errors: if True errors resulting from failed removals will be ignored
    """
    file_pointer = get_upath(file_pointer)
    if ignore_errors:
        try:
            _rmdir_recursive(file_pointer)
        except Exception:  # pylint: disable=broad-except
            # fsspec doesn't have a "ignore_errors" field in the rm method
            pass
    else:
        _rmdir_recursive(file_pointer)


def write_string_to_file(file_pointer: str | Path | UPath, string: str, encoding: str = "utf-8"):
    """Write a string to a text file

    Args:
        file_pointer: file location to write file to
        string: string to write to file
        encoding: Default: 'utf-8', encoding method to write to file with
    """
    file_pointer = get_upath(file_pointer)
    with file_pointer.open("w", encoding=encoding) as _file:
        _file.write(string)


def load_text_file(file_pointer: str | Path | UPath, encoding: str = "utf-8"):
    """Load a text file content to a list of strings.

    Args:
        file_pointer: location of file to read
        encoding: string encoding method used by the file
    Returns:
        text contents of file.
    """
    file_pointer = get_upath(file_pointer)
    with file_pointer.open("r", encoding=encoding) as _text_file:
        text_file = _text_file.readlines()

    return text_file


def load_csv_to_pandas(file_pointer: str | Path | UPath, **kwargs) -> pd.DataFrame:
    """Load a csv file to a pandas dataframe

    Args:
        file_pointer: location of csv file to load
        **kwargs: arguments to pass to pandas `read_csv` loading method
    Returns:
        pandas dataframe loaded from CSV
    """
    file_pointer = get_upath(file_pointer)
    with file_pointer.open("r") as csv_file:
        frame = pd.read_csv(csv_file, **kwargs)
    return frame


def load_csv_to_pandas_generator(
    file_pointer: str | Path | UPath, chunksize=10_000, **kwargs
) -> Generator[pd.DataFrame]:
    """Load a csv file to a pandas dataframe
    Args:
        file_pointer: location of csv file to load
        file_system: fsspec or pyarrow filesystem, default None
        **kwargs: arguments to pass to pandas `read_csv` loading method
    Returns:
        pandas dataframe loaded from CSV
    """
    file_pointer = get_upath(file_pointer)
    with file_pointer.open("r", **kwargs) as csv_file:
        with pd.read_csv(csv_file, chunksize=chunksize, **kwargs) as reader:
            yield from reader


def write_dataframe_to_csv(dataframe: pd.DataFrame, file_pointer: str | Path | UPath, **kwargs):
    """Write a pandas DataFrame to a CSV file

    Args:
        dataframe: DataFrame to write
        file_pointer: location of file to write to
        **kwargs: args to pass to pandas `to_csv` method
    """
    output = dataframe.to_csv(**kwargs)
    write_string_to_file(file_pointer, output)


def write_dataframe_to_parquet(dataframe: pd.DataFrame, file_pointer):
    """Write a pandas DataFrame to a parquet file

    Args:
        dataframe: DataFrame to write
        file_pointer: location of file to write to
    """
    file_pointer = get_upath(file_pointer)
    dataframe.to_parquet(file_pointer.path, filesystem=file_pointer.fs)


def read_parquet_metadata(file_pointer: str | Path | UPath, **kwargs) -> pq.FileMetaData:
    """Read FileMetaData from footer of a single Parquet file.

    Args:
        file_pointer: location of file to read metadata from
        **kwargs: additional arguments to be passed to pyarrow.parquet.read_metadata
    """
    file_pointer = get_upath(file_pointer)
    if file_pointer is None or not file_pointer.exists():
        raise FileNotFoundError("Parquet file does not exist")
    parquet_file = pq.read_metadata(file_pointer.path, filesystem=file_pointer.fs, **kwargs)
    return parquet_file


def read_parquet_dataset(source: str | Path | UPath, **kwargs) -> Tuple[UPath, Dataset]:
    """Read parquet dataset from directory pointer or list of files.

    Note that pyarrow.dataset reads require that directory pointers don't contain a
    leading slash, and the protocol prefix may additionally be removed. As such, we also return
    the directory path that is formatted for pyarrow ingestion for follow-up.

    See more info on source specification and possible kwargs at
    https://arrow.apache.org/docs/python/generated/pyarrow.dataset.dataset.html

    Args:
        source: directory, path, or list of paths to read data from

    Returns:
        Tuple containing a path to the dataset (that is formatted for pyarrow ingestion)
        and the dataset read from disk.
    """
    if pd.api.types.is_list_like(source) and len(source) > 0:
        sample_pointer = source[0]
        sample_pointer = get_upath(sample_pointer)
        file_system = sample_pointer.fs
        source = [str(path) for path in source]
    else:
        source = get_upath(source)
        file_system = source.fs
        source = source.path

    dataset = pds.dataset(
        source,
        filesystem=file_system,
        format="parquet",
        **kwargs,
    )
    return (str(source), dataset)


def write_parquet_metadata(
    schema: Any, file_pointer: str | Path | UPath, metadata_collector: list | None = None, **kwargs
):
    """Write a metadata only parquet file from a schema

    Args:
        schema: schema to be written
        file_pointer: location of file to be written to
        metadata_collector: where to collect metadata information
        **kwargs: additional arguments to be passed to pyarrow.parquet.write_metadata
    """
    file_pointer = get_upath(file_pointer)
    pq.write_metadata(
        schema, file_pointer.path, metadata_collector=metadata_collector, filesystem=file_pointer.fs, **kwargs
    )


def read_fits_image(map_file_pointer: str | Path | UPath):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        file_pointer: location of file to be written
    Returns:
        one-dimensional numpy array of integers where the
        value at each index corresponds to the number of objects found at the healpix pixel.
    """
    map_file_pointer = get_upath(map_file_pointer)
    with tempfile.NamedTemporaryFile() as _tmp_file:
        with map_file_pointer.open("rb") as _map_file:
            map_data = _map_file.read()
            _tmp_file.write(map_data)
            return hp.read_map(_tmp_file.name, nest=True)


def write_fits_image(histogram: np.ndarray, map_file_pointer: str | Path | UPath):
    """Write the object spatial distribution information to a healpix FITS file.

    Args:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        file_pointer: location of file to be written
    """
    map_file_pointer = get_upath(map_file_pointer)
    with tempfile.NamedTemporaryFile() as _tmp_file:
        with map_file_pointer.open("wb") as _map_file:
            hp.write_map(_tmp_file.name, histogram, overwrite=True, dtype=np.int32, nest=True, coord="CEL")
            _map_file.write(_tmp_file.read())


def read_yaml(file_handle: str | Path | UPath):
    """Reads yaml file from filesystem.

    Args:
        file_handle: location of yaml file
    """
    file_handle = get_upath(file_handle)
    with file_handle.open("r", encoding="utf-8") as _file:
        metadata = yaml.safe_load(_file)
    return metadata


def delete_file(file_handle: str | Path | UPath):
    """Deletes file from filesystem.

    Args:
        file_handle: location of file pointer
    """
    file_handle = get_upath(file_handle)
    file_handle.unlink()


def read_parquet_file_to_pandas(
    file_pointer: str | Path | UPath, file_open_kwargs: dict = None, **kwargs
) -> pd.DataFrame:
    """Reads a parquet file to a pandas DataFrame

    Args:
        file_pointer (UPath): File Pointer to a parquet file
        **kwargs: Additional arguments to pass to pandas read_parquet method

    Returns:
        Pandas DataFrame with the data from the parquet file
    """
    file_pointer = get_upath(file_pointer)
    if file_open_kwargs is None:
        file_open_kwargs = {}
    with file_pointer.open("rb", **file_open_kwargs) as parquet_file:
        return pd.read_parquet(parquet_file, **kwargs)



"""Container class to hold catalog metadata and partition iteration"""

from __future__ import annotations

from typing import List, Tuple

import numpy as np

import hats.pixel_math.healpix_shim as hp
from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset
from hats.pixel_math import HealpixPixel
from hats.pixel_math.box_filter import generate_box_moc, wrap_ra_angles
from hats.pixel_math.cone_filter import generate_cone_moc
from hats.pixel_math.polygon_filter import CartesianCoordinates, SphericalCoordinates, generate_polygon_moc
from hats.pixel_math.validators import (
    validate_box_search,
    validate_declination_values,
    validate_polygon,
    validate_radius,
)
from hats.pixel_tree.negative_tree import compute_negative_tree_pixels


class Catalog(HealpixDataset):
    """A HATS Catalog with data stored in a HEALPix Hive partitioned structure

    Catalogs of this type are partitioned spatially, contain `partition_info` metadata specifying
    the pixels in Catalog, and on disk conform to the parquet partitioning structure
    `Norder=/Dir=/Npix=.parquet`
    """

    def filter_by_cone(self, ra: float, dec: float, radius_arcsec: float) -> Catalog:
        """Filter the pixels in the catalog to only include the pixels that overlap with a cone

        Args:
            ra (float): Right Ascension of the center of the cone in degrees
            dec (float): Declination of the center of the cone in degrees
            radius_arcsec (float): Radius of the cone in arcseconds

        Returns:
            A new catalog with only the pixels that overlap with the specified cone
        """
        validate_radius(radius_arcsec)
        validate_declination_values(dec)
        return self.filter_by_moc(generate_cone_moc(ra, dec, radius_arcsec, self.get_max_coverage_order()))

    def filter_by_box(
        self, ra: Tuple[float, float] | None = None, dec: Tuple[float, float] | None = None
    ) -> Catalog:
        """Filter the pixels in the catalog to only include the pixels that overlap with a
        right ascension or declination range. In case both ranges are provided, filtering
        is performed using a polygon.

        Args:
            ra (Tuple[float, float]): Right ascension range, in degrees
            dec (Tuple[float, float]): Declination range, in degrees

        Returns:
            A new catalog with only the pixels that overlap with the specified region
        """
        ra = tuple(wrap_ra_angles(ra)) if ra else None
        validate_box_search(ra, dec)
        return self.filter_by_moc(generate_box_moc(ra, dec, self.get_max_coverage_order()))

    def filter_by_polygon(self, vertices: List[SphericalCoordinates] | List[CartesianCoordinates]) -> Catalog:
        """Filter the pixels in the catalog to only include the pixels that overlap
        with a polygonal sky region.

        Args:
            vertices (List[SphericalCoordinates] | List[CartesianCoordinates]): The vertices
                of the polygon to filter points with, in lists of (ra,dec) or (x,y,z) points
                on the unit sphere.

        Returns:
            A new catalog with only the pixels that overlap with the specified polygon.
        """
        if all(len(vertex) == 2 for vertex in vertices):
            ra, dec = np.array(vertices).T
            validate_declination_values(dec)
            # Get the coordinates vector on the unit sphere if we were provided
            # with polygon spherical coordinates of ra and dec
            cart_vertices = hp.ang2vec(ra, dec, lonlat=True)
        else:
            cart_vertices = vertices
        validate_polygon(cart_vertices)
        return self.filter_by_moc(generate_polygon_moc(cart_vertices, self.get_max_coverage_order()))

    def generate_negative_tree_pixels(self) -> List[HealpixPixel]:
        """Get the leaf nodes at each healpix order that have zero catalog data.

        For example, if an example catalog only had data points in pixel 0 at
        order 0, then this method would return order 0's pixels 1 through 11.
        Used for getting full coverage on margin caches.

        Returns:
            List of HealpixPixels representing the 'negative tree' for the catalog.
        """
        return compute_negative_tree_pixels(self.pixel_tree)



from __future__ import annotations

from pathlib import Path
from typing import List

from upath import UPath


def get_upath(path: str | Path | UPath) -> UPath:
    """Returns a file pointer from a path string"""
    if not path:
        return None
    if isinstance(path, UPath):
        return path
    return UPath(path)


def append_paths_to_pointer(pointer: str | Path | UPath, *paths: str) -> UPath:
    """Append directories and/or a file name to a specified file pointer.

    Args:
        pointer: `FilePointer` object to add path to
        paths: any number of directory names optionally followed by a file name to append to the
            pointer

    Returns:
        New file pointer to path given by joining given pointer and path names
    """
    pointer = get_upath(pointer)
    return pointer.joinpath(*paths)


def does_file_or_directory_exist(pointer: str | Path | UPath) -> bool:
    """Checks if a file or directory exists for a given file pointer

    Args:
        pointer: File Pointer to check if file or directory exists at

    Returns:
        True if file or directory at `pointer` exists, False if not
    """
    pointer = get_upath(pointer)
    return pointer.exists()


def is_regular_file(pointer: str | Path | UPath) -> bool:
    """Checks if a regular file (NOT a directory) exists for a given file pointer.

    Args:
        pointer: File Pointer to check if a regular file

    Returns:
        True if regular file at `pointer` exists, False if not or is a directory
    """
    pointer = get_upath(pointer)
    return pointer.is_file()


def find_files_matching_path(pointer: str | Path | UPath, *paths: str) -> List[UPath]:
    """Find files or directories matching the provided path parts.

    Args:
        pointer: base File Pointer in which to find contents
        paths: any number of directory names optionally followed by a file name.
            directory or file names may be replaced with `*` as a matcher.
    Returns:
        New file pointers to files found matching the path
    """
    pointer = get_upath(pointer)

    if len(paths) == 0:
        return [pointer]

    matcher = pointer.fs.sep.join(paths)
    contents = []
    for child in pointer.rglob(matcher):
        contents.append(child)

    if len(contents) == 0:
        return []

    contents.sort()
    return contents


def directory_has_contents(pointer: str | Path | UPath) -> bool:
    """Checks if a directory already has some contents (any files or subdirectories)

    Args:
        pointer: File Pointer to check for existing contents

    Returns:
        True if there are any files or subdirectories below this directory.
    """
    pointer = get_upath(pointer)
    return len(find_files_matching_path(pointer, "*")) > 0


def get_directory_contents(pointer: str | Path | UPath) -> List[UPath]:
    """Finds all files and directories in the specified directory.

    NB: This is not recursive, and will return only the first level of directory contents.

    Args:
        pointer: File Pointer in which to find contents

    Returns:
        New file pointers to files or subdirectories below this directory.
    """
    pointer = get_upath(pointer)
    contents = []
    for child in pointer.iterdir():
        contents.append(child)

    if len(contents) == 0:
        return []

    contents.sort()
    return contents



import pytest

from hats.loaders import read_hats
from hats.pixel_math import HealpixPixel

# pylint: disable= redefined-outer-name


@pytest.fixture
def small_sky_catalog(small_sky_dir):
    return read_hats(small_sky_dir)


@pytest.fixture()
def small_sky_pixels():
    return [
        HealpixPixel(0, 11),
    ]


@pytest.fixture
def small_sky_order1_catalog(small_sky_order1_dir):
    return read_hats(small_sky_order1_dir)


@pytest.fixture()
def small_sky_order1_pixels():
    return [
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]


@pytest.fixture
def pixel_list_depth_first():
    """A list of pixels that are sorted by Norder (depth-first)"""
    # pylint: disable=duplicate-code
    return [
        HealpixPixel(0, 10),
        HealpixPixel(1, 33),
        HealpixPixel(1, 35),
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(2, 128),
        HealpixPixel(2, 130),
        HealpixPixel(2, 131),
    ]


@pytest.fixture
def pixel_list_breadth_first():
    """The same pixels in the above `pixel_list_depth_first` list, but
    in breadth-first traversal order by the healpix pixel hierarchy.

    See tree for illustration (brackets indicate inner node)::

        .
        ├── <0,8>
        │   ├── <1,32>
        │   │   ├── 2,128
        │   │   ├── 2,130
        │   │   └── 2,131
        │   ├── 1,33
        │   └── 1,35
        ├── 0,10
        └── <0,11>
            ├── 1,44
            ├── 1,45
            └── 1,46
    """
    return [
        HealpixPixel(2, 128),
        HealpixPixel(2, 130),
        HealpixPixel(2, 131),
        HealpixPixel(1, 33),
        HealpixPixel(1, 35),
        HealpixPixel(0, 10),
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
    ]



import astropy.units as u
from mocpy import MOC


def generate_cone_moc(ra: float, dec: float, radius_arcsec: float, order: int) -> MOC:
    """Generate a MOC object that covers the cone"""
    return MOC.from_cone(lon=ra * u.deg, lat=dec * u.deg, radius=radius_arcsec * u.arcsec, max_depth=order)



from __future__ import annotations

from typing import Tuple, Union

from hats.pixel_math.healpix_pixel import HealpixPixel

HealpixInputTypes = Union[HealpixPixel, Tuple[int, int]]


def get_healpix_pixel(pixel: HealpixInputTypes) -> HealpixPixel:
    """Function to convert argument of either HealpixPixel or a tuple of (order, pixel) to a
    HealpixPixel

    Args:
        pixel: an object to be converted to a HealpixPixel object
    """

    if isinstance(pixel, tuple):
        if len(pixel) != 2:
            raise ValueError("Tuple must contain two values: HEALPix order and HEALPix pixel number")
        return HealpixPixel(order=pixel[0], pixel=pixel[1])
    if isinstance(pixel, HealpixPixel):
        return pixel
    raise TypeError("pixel must either be of type `HealpixPixel` or tuple (order, pixel)")


def get_healpix_tuple(pixel: HealpixInputTypes) -> Tuple[int, int]:
    """Function to convert argument of either HealpixPixel or a tuple of (order, pixel) to a
    tuple of (order, pixel)

    Args:
        pixel: an object to be converted to a HealpixPixel object
    """

    if isinstance(pixel, tuple):
        if len(pixel) != 2:
            raise ValueError("Tuple must contain two values: HEALPix order and HEALPix pixel number")
        return pixel
    if isinstance(pixel, HealpixPixel):
        return (pixel.order, pixel.pixel)
    raise TypeError("pixel must either be of type `HealpixPixel` or tuple (order, pixel)")



import healpy as hp
import numpy as np

# pylint: disable=missing-function-docstring

## Arithmetic conversions


def nside2order(param):
    return hp.nside2order(param)


def order2nside(param):
    return hp.order2nside(param)


def npix2nside(param):
    return hp.npix2nside(param)


def nside2npix(param):
    return hp.nside2npix(param)


def order2npix(param):
    return hp.order2npix(param)


def npix2order(param):
    return hp.npix2order(param)


def nside2resol(*args, **kwargs):
    return hp.nside2resol(*args, **kwargs)


def nside2pixarea(*args, **kwargs):
    return hp.nside2pixarea(*args, **kwargs)


def ang2pix(*args, **kwargs):
    return hp.ang2pix(*args, **kwargs)


def ring2nest(*args, **kwargs):
    return hp.ring2nest(*args, **kwargs)


def unseen_pixel():
    return hp.pixelfunc.UNSEEN


## Query


def query_strip(*args, **kwargs):
    return hp.query_strip(*args, **kwargs)


def query_polygon(*args, **kwargs):
    return hp.query_polygon(*args, **kwargs)


def boundaries(*args, **kwargs):
    return hp.boundaries(*args, **kwargs)


def get_all_neighbours(*args, **kwargs):
    return hp.get_all_neighbours(*args, **kwargs)


## Coordinate conversion


def ang2vec(*args, **kwargs):
    return hp.ang2vec(*args, **kwargs)


def pix2ang(*args, **kwargs):
    return hp.pix2ang(*args, **kwargs)


def vec2dir(*args, **kwargs):
    return hp.vec2dir(*args, **kwargs)


def pix2xyf(*args, **kwargs):
    return hp.pix2xyf(*args, **kwargs)


## FITS
def read_map(*args, **kwargs):
    return hp.read_map(*args, **kwargs)


def write_map(*args, **kwargs):
    return hp.write_map(*args, **kwargs)


## Custom functions


def avgsize2mindist(avg_size: np.ndarray) -> np.ndarray:
    """Get the minimum distance between pixels for a given average size

    Average size is a "resolution" in healpy terms, i.e. given by
    `healpy.nside2resol`.

    We don't have the precise geometry of the healpix grid yet,
    so we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    avg_size : np.ndarray of float
        The average size of a healpix pixel

    Returns
    -------
    np.ndarray of float
        The minimum distance between pixels for the given average size
    """
    return avg_size / 1.6


def mindist2avgsize(mindist: np.ndarray) -> np.ndarray:
    """Get the average size for a given minimum distance between pixels

    Average size is a "resolution" in healpy terms, i.e. given by
    `healpy.nside2resol`.

    We don't have the precise geometry of the healpix grid yet,
    so we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    mindist : np.ndarray of float
        The minimum distance between pixels

    Returns
    -------
    np.ndarray of float
        The average size of a healpix pixel for the given minimum distance
        between pixels.
    """
    return mindist * 1.6


def avgsize2order(avg_size_arcmin: np.ndarray) -> np.ndarray:
    """Get the largest order with average healpix size larger than avg_size_arcmin

    "Average" size is healpy's "resolution", so this function is
    reverse to healpy.nside2resol(healpy.order2nside(order))..

    Parameters
    ----------
    avg_size_arcmin : np.ndarray of float
        The average size of a healpix pixel in arcminutes

    Returns
    -------
    np.ndarray of int
        The largest healpix order for which the average size is larger than avg_size_arcmin
    """
    # resolution = sqrt(4pi / (12 * 4**order)) => order = log2(sqrt(4pi / 12) / resolution)
    order_float = np.log2(np.sqrt(np.pi / 3) / np.radians(avg_size_arcmin / 60.0))
    return np.array(order_float, dtype=int)


def margin2order(margin_thr_arcmin: np.ndarray) -> np.ndarray:
    """Get the largest order for which distance between pixels is less than margin_thr_arcmin

    We don't have the precise geometry of the healpix grid yet,
    we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    margin_thr_arcmin : np.ndarray of float
        The minimum distance between pixels in arcminutes

    Returns
    -------
    np.ndarray of int
        The largest healpix order for which the distance between pixels is less than margin_thr_arcmin
    """
    avg_size_arcmin = mindist2avgsize(margin_thr_arcmin)
    return avgsize2order(avg_size_arcmin)



from __future__ import annotations

import warnings
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from mocpy import MOC
from typing_extensions import Self
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.catalog.dataset import Dataset
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.partition_info import PartitionInfo
from hats.io import file_io, paths
from hats.io.file_io import read_parquet_metadata
from hats.pixel_math import HealpixPixel
from hats.pixel_tree import PixelAlignment, PixelAlignmentType
from hats.pixel_tree.moc_filter import filter_by_moc
from hats.pixel_tree.pixel_alignment import align_with_mocs
from hats.pixel_tree.pixel_tree import PixelTree

PixelInputTypes = Union[PartitionInfo, PixelTree, List[HealpixPixel]]


class HealpixDataset(Dataset):
    """A HATS dataset partitioned with a HEALPix partitioning structure.

    Catalogs of this type are partitioned based on the ra and dec of the points with each partition
    containing points within a given HEALPix pixel. The files are in the form::

        Norder=/Dir=/Npix=.parquet
    """

    def __init__(
        self,
        catalog_info: TableProperties,
        pixels: PixelInputTypes,
        catalog_path: str | Path | UPath | None = None,
        moc: MOC | None = None,
        schema: pa.Schema | None = None,
    ) -> None:
        """Initializes a Catalog

        Args:
            catalog_info: TableProperties object with catalog metadata
            pixels: Specifies the pixels contained in the catalog. Can be either a
                list of HealpixPixel, `PartitionInfo object`, or a `PixelTree` object
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
            moc (mocpy.MOC): MOC object representing the coverage of the catalog
            schema (pa.Schema): The pyarrow schema for the catalog
        """
        super().__init__(catalog_info, catalog_path=catalog_path)
        self.partition_info = self._get_partition_info_from_pixels(pixels)
        self.pixel_tree = self._get_pixel_tree_from_pixels(pixels)
        self.moc = moc
        self.schema = schema

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Get healpix pixel objects for all pixels contained in the catalog.

        Returns:
            List of HealpixPixel
        """
        return self.partition_info.get_healpix_pixels()

    @staticmethod
    def _get_partition_info_from_pixels(pixels: PixelInputTypes) -> PartitionInfo:
        if isinstance(pixels, PartitionInfo):
            return pixels
        if isinstance(pixels, PixelTree):
            return PartitionInfo.from_healpix(pixels.get_healpix_pixels())
        if pd.api.types.is_list_like(pixels):
            return PartitionInfo.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    @staticmethod
    def _get_pixel_tree_from_pixels(pixels: PixelInputTypes) -> PixelTree:
        if isinstance(pixels, PartitionInfo):
            return PixelTree.from_healpix(pixels.get_healpix_pixels())
        if isinstance(pixels, PixelTree):
            return pixels
        if pd.api.types.is_list_like(pixels):
            return PixelTree.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    @classmethod
    def _read_args(cls, catalog_base_dir: str | Path | UPath) -> Tuple[TableProperties, PartitionInfo]:
        args = super()._read_args(catalog_base_dir)
        partition_info = PartitionInfo.read_from_dir(catalog_base_dir)
        return args + (partition_info,)

    @classmethod
    def _read_kwargs(cls, catalog_base_dir: str | Path | UPath) -> dict:
        kwargs = super()._read_kwargs(catalog_base_dir)
        kwargs["moc"] = cls._read_moc_from_point_map(catalog_base_dir)
        kwargs["schema"] = cls._read_schema_from_metadata(catalog_base_dir)
        return kwargs

    @classmethod
    def _read_moc_from_point_map(cls, catalog_base_dir: str | Path | UPath) -> MOC | None:
        """Reads a MOC object from the `point_map.fits` file if it exists in the catalog directory"""
        point_map_path = paths.get_point_map_file_pointer(catalog_base_dir)
        if not file_io.does_file_or_directory_exist(point_map_path):
            return None
        fits_image = file_io.read_fits_image(point_map_path)
        order = hp.nside2order(hp.npix2nside(len(fits_image)))
        boolean_skymap = fits_image.astype(bool)
        ipix = np.where(boolean_skymap)[0]
        orders = np.full(ipix.shape, order)
        return MOC.from_healpix_cells(ipix, orders, order)

    @classmethod
    def _read_schema_from_metadata(cls, catalog_base_dir: str | Path | UPath) -> pa.Schema | None:
        """Reads the schema information stored in the _common_metadata or _metadata files."""
        common_metadata_file = paths.get_common_metadata_pointer(catalog_base_dir)
        common_metadata_exists = file_io.does_file_or_directory_exist(common_metadata_file)
        metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
        metadata_exists = file_io.does_file_or_directory_exist(metadata_file)
        if not (common_metadata_exists or metadata_exists):
            warnings.warn(
                "_common_metadata or _metadata files not found for this catalog."
                "The arrow schema will not be set."
            )
            return None
        schema_file = common_metadata_file if common_metadata_exists else metadata_file
        metadata = read_parquet_metadata(schema_file)
        return metadata.schema.to_arrow_schema()

    @classmethod
    def _check_files_exist(cls, catalog_base_dir: str | Path | UPath):
        super()._check_files_exist(catalog_base_dir)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
        if not (
            file_io.does_file_or_directory_exist(partition_info_file)
            or file_io.does_file_or_directory_exist(metadata_file)
        ):
            raise FileNotFoundError(
                f"_metadata or partition info file is required in catalog directory {catalog_base_dir}"
            )

    def get_max_coverage_order(self) -> int:
        """Gets the maximum HEALPix order for which the coverage of the catalog is known from the pixel
        tree and moc if it exists"""
        max_order = (
            max(self.moc.max_order, self.pixel_tree.get_max_depth())
            if self.moc is not None
            else self.pixel_tree.get_max_depth()
        )
        return max_order

    def filter_from_pixel_list(self, pixels: List[HealpixPixel]) -> Self:
        """Filter the pixels in the catalog to only include any that overlap with the requested pixels.

        Args:
            pixels (List[HealpixPixels]): the pixels to include

        Returns:
            A new catalog with only the pixels that overlap with the given pixels. Note that we reset the
            total_rows to None, as updating would require a scan over the new pixel sizes.
        """
        orders = np.array([p.order for p in pixels])
        pixel_inds = np.array([p.pixel for p in pixels])
        max_order = np.max(orders) if len(orders) > 0 else 0
        moc = MOC.from_healpix_cells(ipix=pixel_inds, depth=orders, max_depth=max_order)
        return self.filter_by_moc(moc)

    def filter_by_moc(self, moc: MOC) -> Self:
        """Filter the pixels in the catalog to only include the pixels that overlap with the moc provided.

        Args:
            moc (mocpy.MOC): the moc to filter by

        Returns:
            A new catalog with only the pixels that overlap with the moc. Note that we reset the total_rows
        to 0, as updating would require a scan over the new pixel sizes."""
        filtered_tree = filter_by_moc(self.pixel_tree, moc)
        filtered_moc = self.moc.intersection(moc) if self.moc is not None else None
        filtered_catalog_info = self.catalog_info.copy_and_update(total_rows=0)
        return self.__class__(filtered_catalog_info, filtered_tree, moc=filtered_moc, schema=self.schema)

    def align(
        self, other_cat: Self, alignment_type: PixelAlignmentType = PixelAlignmentType.INNER
    ) -> PixelAlignment:
        """Performs an alignment to another catalog, using the pixel tree and mocs if available

        An alignment compares the pixel structures of the two catalogs, checking which pixels overlap.
        The alignment includes the mapping of all pairs of pixels in each tree that overlap with each other,
        and the aligned tree which consists of the overlapping pixels in the two input catalogs, using the
        higher order pixels where there is overlap with differing orders.

        For more information, see this document:
        https://docs.google.com/document/d/1gqb8qb3HiEhLGNav55LKKFlNjuusBIsDW7FdTkc5mJU/edit?usp=sharing

        Args:
            other_cat (Catalog): The catalog to align to
            alignment_type (PixelAlignmentType): The type of alignment describing how to handle nodes which
            exist in one tree but not the other. Mirrors the 'how' argument of a pandas/sql join. Options are:

                - "inner" - only use pixels that appear in both catalogs
                - "left" - use all pixels that appear in the left catalog and any overlapping from the right
                - "right" - use all pixels that appear in the right catalog and any overlapping from the left
                - "outer" - use all pixels from both catalogs

        Returns (PixelAlignment):
            A `PixelAlignment` object with the alignment from the two catalogs
        """
        return align_with_mocs(
            self.pixel_tree, other_cat.pixel_tree, self.moc, other_cat.moc, alignment_type=alignment_type
        )



"""Test construction (and de-construction) of the healpix-based spatial index"""

import numpy as np
import numpy.testing as npt
import pytest

import hats.pixel_math.healpix_shim as hp
from hats.pixel_math.spatial_index import (
    SPATIAL_INDEX_ORDER,
    compute_spatial_index,
    healpix_to_spatial_index,
    spatial_index_to_healpix,
)


def test_single():
    """Single point. Adheres to specification."""
    result = compute_spatial_index([5], [5])
    expected = hp.ang2pix(
        2**SPATIAL_INDEX_ORDER,
        [5],
        [5],
        nest=True,
        lonlat=True,
    )

    npt.assert_array_equal(result, expected)


def test_jagged_list():
    """Arrays of mismatched lengths."""
    with pytest.raises(ValueError):
        compute_spatial_index([5, 1, 5], [5])


def test_short_list():
    """Multiple points that will sit in the same higher-order-pixel."""
    ra = [5, 1, 5]
    dec = [5, 1, 5]
    result = compute_spatial_index(ra, dec)
    expected = hp.ang2pix(
        2**SPATIAL_INDEX_ORDER,
        ra,
        dec,
        nest=True,
        lonlat=True,
    )
    npt.assert_array_equal(result, expected)


def test_list():
    """Multiple points that will sit in the same higher-order-pixel."""
    ra = [5, 5, 5, 1, 5, 5, 5, 1, 5]
    dec = [5, 5, 5, 1, 5, 5, 5, 1, 5]
    result = compute_spatial_index(ra, dec)
    expected = hp.ang2pix(
        2**SPATIAL_INDEX_ORDER,
        ra,
        dec,
        nest=True,
        lonlat=True,
    )
    npt.assert_array_equal(result, expected)


def test_load():
    """Generate a kinda big array and make sure the method completes in under a second.
    If this method is failing due to timeouts, please refactor to keep within the time limit.
    """
    rng = np.random.default_rng(seed=800)
    test_num = 1_000_000

    ra_arr = rng.random(test_num)
    dec_arr = rng.random(test_num)
    result = compute_spatial_index(ra_arr, dec_arr)

    assert len(result) == test_num


def test_spatial_index_to_healpix():
    """Test the inverse operation"""
    ids = [
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
    ]

    result = spatial_index_to_healpix(ids)

    expected = [
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
    ]

    npt.assert_array_equal(result, expected)


def test_spatial_index_to_healpix_low_order():
    """Test the inverse operation"""
    ids = [
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
        3458764513820540924,
        3458764513820540924,
        3138264513820540924,  # out of sequence
        3458764513820540924,
    ]

    result = spatial_index_to_healpix(ids, target_order=4)

    expected = [i >> (2 * (29 - 4)) for i in ids]

    npt.assert_array_equal(result, expected)


def test_healpix_to_spatial_index_single():
    orders = [3, 3, 4, 1]
    pixels = [0, 12, 1231, 11]
    pixels_at_high_order = [p * (4 ** (SPATIAL_INDEX_ORDER - o)) for o, p in zip(orders, pixels)]
    lon, lat = hp.pix2ang(
        [2**SPATIAL_INDEX_ORDER] * len(orders), pixels_at_high_order, nest=True, lonlat=True
    )
    actual_spatial_indices = compute_spatial_index(lon, lat)
    test_spatial_indices = [healpix_to_spatial_index(o, p) for o, p in zip(orders, pixels)]
    assert np.all(test_spatial_indices == actual_spatial_indices)


def test_healpix_to_spatial_index_array():
    orders = [3, 3, 4, 1]
    pixels = [0, 12, 1231, 11]
    pixels_at_high_order = [p * (4 ** (SPATIAL_INDEX_ORDER - o)) for o, p in zip(orders, pixels)]
    lon, lat = hp.pix2ang(
        [2**SPATIAL_INDEX_ORDER] * len(orders), pixels_at_high_order, nest=True, lonlat=True
    )
    actual_spatial_indices = compute_spatial_index(lon, lat)
    test_spatial_indices = healpix_to_spatial_index(orders, pixels)
    assert np.all(test_spatial_indices == actual_spatial_indices)



import numba
import numpy as np
from mocpy import MOC
from numba import njit

from hats.pixel_tree.pixel_tree import PixelTree


def filter_by_moc(
    tree: PixelTree,
    moc: MOC,
) -> PixelTree:
    """Filters a pixel tree to only include the pixels that overlap with the pixels in the moc

    Args:
        tree (PixelTree): The tree to perform the filtering on
        moc (mocpy.MOC): The moc to use to filter

    Returns:
        A new PixelTree object with only the pixels from the input tree that overlap with the moc.
    """
    moc_ranges = moc.to_depth29_ranges
    # Convert tree intervals to order 29 to match moc intervals
    tree_29_ranges = tree.tree << (2 * (29 - tree.tree_order))
    tree_mask = perform_filter_by_moc(tree_29_ranges, moc_ranges)
    return PixelTree(tree.tree[tree_mask], tree.tree_order)


@njit(
    numba.bool_[::1](
        numba.int64[:, :],
        numba.uint64[:, :],
    )
)
def perform_filter_by_moc(
    tree: np.ndarray,
    moc: np.ndarray,
) -> np.ndarray:  # pragma: no cover
    """Performs filtering with lists of pixel intervals

    Input interval lists must be at the same order.

    Args:
        tree (np.ndarray): Array of pixel intervals to be filtered
        moc (np.ndarray): Array of pixel intervals to be used to filter

    Returns:
        A boolean array of dimension tree.shape[0] which masks which pixels in tree overlap with the pixels in
        moc
    """
    output = np.full(tree.shape[0], fill_value=False, dtype=np.bool_)
    tree_index = 0
    moc_index = 0
    while tree_index < len(tree) and moc_index < len(moc):
        tree_pix = tree[tree_index]
        moc_pix = moc[moc_index]
        if tree_pix[0] >= moc_pix[1]:
            # Don't overlap, tree pixel ahead so move onto next MOC pixel
            moc_index += 1
            continue
        if moc_pix[0] >= tree_pix[1]:
            # Don't overlap, MOC pixel ahead so move onto next tree pixel
            tree_index += 1
            continue
        # Pixels overlap, so include current tree pixel and check next tree pixel
        output[tree_index] = True
        tree_index += 1
    return output



import pytest

from hats.catalog.dataset.table_properties import TableProperties


@pytest.mark.parametrize("data_dir", ["catalog", "dataset", "index_catalog", "margin_cache"])
def test_read_from_file_round_trip(test_data_dir, data_dir, tmp_path):
    dataset_path = test_data_dir / "info_only" / data_dir

    table_properties = TableProperties.read_from_dir(dataset_path)
    table_properties.to_properties_file(tmp_path)
    round_trip_properties = TableProperties.read_from_dir(tmp_path)

    assert table_properties == round_trip_properties

    kwarg_properties = TableProperties(**round_trip_properties.model_dump(by_alias=False, exclude_none=True))
    assert table_properties == kwarg_properties


def test_properties_parsing():
    table_properties = TableProperties(
        catalog_name="foo",
        catalog_type="index",
        total_rows=15,
        extra_columns="a , b",
        indexing_column="a",
        primary_catalog="bar",
        hats_copyright="LINCC Frameworks 2024",
    )
    assert table_properties.extra_columns == ["a", "b"]

    # hats_copyright is not part of the named args, so it shouldn't show up in the debug string
    assert (
        str(table_properties)
        == """  catalog_name foo
  catalog_type index
  total_rows 15
  primary_catalog bar
  indexing_column a
  extra_columns a b
"""
    )
    table_properties_using_list = TableProperties(
        catalog_name="foo",
        catalog_type="index",
        total_rows=15,
        extra_columns=["a", "b"],
        indexing_column="a",
        primary_catalog="bar",
        hats_copyright="LINCC Frameworks 2024",
    )
    assert table_properties_using_list == table_properties


def test_properties_allowed_required():
    # Missing required field indexing_column
    with pytest.raises(ValueError, match="indexing_column"):
        TableProperties(
            catalog_name="foo",
            catalog_type="index",
            total_rows=15,
            primary_catalog="bar",
        )

    # join_column is only allowed on association catalogs
    with pytest.raises(ValueError, match="join_column"):
        TableProperties(
            catalog_name="foo",
            catalog_type="index",
            total_rows=15,
            indexing_column="a",
            primary_catalog="bar",
            join_column="b",
        )

    # extra_columnsss is a typo
    with pytest.raises(ValueError, match="extra_columnsss"):
        TableProperties(
            catalog_name="foo",
            catalog_type="index",
            total_rows=15,
            indexing_column="a",
            primary_catalog="bar",
            extra_columnsss=["beep"],
        )


def test_copy_and_update():
    initital_properties = TableProperties(
        catalog_name="foo",
        catalog_type="index",
        total_rows=15,
        indexing_column="a",
        primary_catalog="bar",
    )
    prop_a = initital_properties.copy_and_update()
    assert initital_properties == prop_a

    prop_b = initital_properties.copy_and_update(catalog_name="bar")
    assert initital_properties != prop_b
    assert prop_b.catalog_name == "bar"

    prop_d = initital_properties.copy_and_update(**{"catalog_name": "bar"})
    assert initital_properties != prop_d
    assert prop_d.catalog_name == "bar"
    assert prop_b == prop_d

    prop_c = initital_properties.copy_and_update(moc_sky_fraction=0.54)
    assert initital_properties != prop_c
    assert prop_c.__pydantic_extra__["moc_sky_fraction"] == pytest.approx(0.54)

    # extra_columnsss is a typo
    with pytest.raises(ValueError, match="extra_columnsss"):
        initital_properties.copy_and_update(extra_columnsss=0.54)

    with pytest.raises(ValueError, match="extra_columnsss"):
        initital_properties.copy_and_update(**{"extra_columnsss": 0.54})


def test_read_from_dir_branches(
    small_sky_dir,
    small_sky_order1_dir,
    association_catalog_path,
    small_sky_source_object_index_dir,
    margin_catalog_path,
    small_sky_source_dir,
):
    TableProperties.read_from_dir(small_sky_dir)
    TableProperties.read_from_dir(small_sky_order1_dir)
    TableProperties.read_from_dir(association_catalog_path)
    TableProperties.read_from_dir(small_sky_source_object_index_dir)
    TableProperties.read_from_dir(margin_catalog_path)
    TableProperties.read_from_dir(small_sky_source_dir)



from __future__ import annotations

from typing import Callable, Dict, List

import numba
import numpy as np
import pandas as pd
from mocpy import MOC
from numba import njit

from hats.pixel_math.healpix_pixel_function import get_pixels_from_intervals
from hats.pixel_tree.moc_filter import perform_filter_by_moc
from hats.pixel_tree.pixel_alignment_types import PixelAlignmentType
from hats.pixel_tree.pixel_tree import PixelTree

LEFT_INCLUDE_ALIGNMENT_TYPES = [PixelAlignmentType.LEFT, PixelAlignmentType.OUTER]
RIGHT_INCLUDE_ALIGNMENT_TYPES = [PixelAlignmentType.RIGHT, PixelAlignmentType.OUTER]


NONE_PIX = np.array([-1, -1])
LEFT_SIDE = True
RIGHT_SIDE = False


# pylint: disable=R0903
class PixelAlignment:
    """Represents how two pixel trees align with each other, meaning which pixels match
    or overlap between the catalogs, and a new tree with the smallest pixels from each tree

    For more information on the pixel alignment algorithm, view this document:
    https://docs.google.com/document/d/1gqb8qb3HiEhLGNav55LKKFlNjuusBIsDW7FdTkc5mJU/edit?usp=sharing

    Attributes:
        pixel_mapping: A dataframe where each row contains a pixel from each tree that match, and
            which pixel in the aligned tree they match with
        pixel_tree: The aligned tree generated by using the smallest pixels in each tree. For
            example, a tree with pixels at order 0, pixel 1, and a tree with order 1, pixel 4,5,6,
            and 7, would result in the smaller order 1 pixels in the aligned tree.
        alignment_type: The type of alignment describing how to handle nodes which exist in one tree
            but not the other. Options are:

                - inner - only use pixels that appear in both catalogs
                - left - use all pixels that appear in the left catalog and any overlapping from the right
                - right - use all pixels that appear in the right catalog and any overlapping from the left
                - outer - use all pixels from both catalogs
    """

    PRIMARY_ORDER_COLUMN_NAME = "primary_Norder"
    PRIMARY_PIXEL_COLUMN_NAME = "primary_Npix"
    JOIN_ORDER_COLUMN_NAME = "join_Norder"
    JOIN_PIXEL_COLUMN_NAME = "join_Npix"
    ALIGNED_ORDER_COLUMN_NAME = "aligned_Norder"
    ALIGNED_PIXEL_COLUMN_NAME = "aligned_Npix"

    def __init__(
        self,
        aligned_tree: PixelTree,
        pixel_mapping: pd.DataFrame,
        alignment_type: PixelAlignmentType,
    ) -> None:
        self.pixel_tree = aligned_tree
        self.pixel_mapping = pixel_mapping
        self.alignment_type = alignment_type


def align_trees(
    left: PixelTree, right: PixelTree, alignment_type: PixelAlignmentType = PixelAlignmentType.INNER
) -> PixelAlignment:
    """Generate a `PixelAlignment` object from two pixel trees

    A `PixelAlignment` represents how two pixel trees align with each other, meaning which pixels
    match or overlap between the catalogs, and includes a new tree with the smallest pixels from
    each tree

    For more information on the pixel alignment algorithm, view this document:
    https://docs.google.com/document/d/1gqb8qb3HiEhLGNav55LKKFlNjuusBIsDW7FdTkc5mJU/edit?usp=sharing

    Args:
        left: The left tree to align
        right: The right tree to align
        alignment_type: The type of alignment describing how to handle nodes which exist in one tree
            but not the other. Options are:

                - inner - only use pixels that appear in both catalogs
                - left - use all pixels that appear in the left catalog and any overlapping from the right
                - right - use all pixels that appear in the right catalog and any overlapping from the left
                - outer - use all pixels from both catalogs

    Returns:
        The `PixelAlignment` object with the alignment from the two trees
    """
    max_n = max(left.tree_order, right.tree_order)

    # Shifts left and right intervals to the same order
    left_aligned = left.tree << (2 * (max_n - left.tree_order))
    right_aligned = right.tree << (2 * (max_n - right.tree_order))

    if alignment_type == PixelAlignmentType.INNER:
        mapping = perform_inner_align_trees(left_aligned, right_aligned)
    else:
        include_all_left = alignment_type in LEFT_INCLUDE_ALIGNMENT_TYPES
        include_all_right = alignment_type in RIGHT_INCLUDE_ALIGNMENT_TYPES
        mapping = perform_align_trees(left_aligned, right_aligned, include_all_left, include_all_right)
        mapping = np.array(mapping).T
    result_tree = mapping[4:6].T if len(mapping) > 0 else np.empty((0, 2), dtype=np.int64)
    result_mapping = get_pixel_mapping_df(mapping, max_n)
    return PixelAlignment(PixelTree(result_tree, max_n), result_mapping, alignment_type)


def get_pixel_mapping_df(mapping: np.ndarray, map_order: int) -> pd.DataFrame:
    """Construct a DataFrame with HEALPix orders and pixels mapping left right and aligned pixels

    Args:
        mapping (np.ndarray): array of shape (6, len(aligned_pixels)) where the first two rows are the
            intervals for the left pixels, the next two for right pixels, and the last two for aligned pixels
        map_order (int): The HEALPix order of the intervals in the mapping array

    Returns:
        A DataFrame with the orders and pixels of the aligned left and right pixels,
    """
    if len(mapping) > 0:
        l_orders, l_pixels = get_pixels_from_intervals(mapping[0:2].T, map_order).T
        r_orders, r_pixels = get_pixels_from_intervals(mapping[2:4].T, map_order).T
        a_orders, a_pixels = get_pixels_from_intervals(mapping[4:6].T, map_order).T
    else:
        l_orders, l_pixels, r_orders, r_pixels, a_orders, a_pixels = [], [], [], [], [], []
    result_mapping = pd.DataFrame.from_dict(
        {
            PixelAlignment.PRIMARY_ORDER_COLUMN_NAME: l_orders,
            PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME: l_pixels,
            PixelAlignment.JOIN_ORDER_COLUMN_NAME: r_orders,
            PixelAlignment.JOIN_PIXEL_COLUMN_NAME: r_pixels,
            PixelAlignment.ALIGNED_ORDER_COLUMN_NAME: a_orders,
            PixelAlignment.ALIGNED_PIXEL_COLUMN_NAME: a_pixels,
        }
    )
    result_mapping.replace(-1, None, inplace=True)
    return result_mapping


# pylint: disable=too-many-statements
@njit(numba.int64[::1, :](numba.int64[:, :], numba.int64[:, :]))
def perform_inner_align_trees(
    left: np.ndarray,
    right: np.ndarray,
) -> np.ndarray:
    """Performs an inner alignment on arrays of pixel intervals

    Pixel interval lists must be of to the same order

    Args:
        left (np.ndarray): the left array of intervals
        right (np.ndarray): the right array of intervals

    Returns (List[np.ndarray]):
        The pixel mapping of the matching left, right, and aligned pixels with each row containing an array of
        [left_order, left_pixel, right_order, right_pixel, aligned_order, aligned_pixel]
    """
    max_out_len = left.shape[0] + right.shape[0]
    mapping = np.zeros((max_out_len, 6), dtype=np.int64)
    left_index = 0
    right_index = 0
    out_index = 0
    while left_index < len(left) and right_index < len(right):
        left_pix = left[left_index]
        right_pix = right[right_index]
        if left_pix[0] >= right_pix[1]:
            # left pix ahead of right, no overlap, so move right on
            right_index += 1
            continue
        if right_pix[0] >= left_pix[1]:
            # right pix ahead of left, no overlap, so move left on
            left_index += 1
            continue
        left_size = left_pix[1] - left_pix[0]
        right_size = right_pix[1] - right_pix[0]
        if left_size == right_size:
            # overlapping & same size => same pixel so add and move both on
            mapping[out_index][0:2], mapping[out_index][2:4], mapping[out_index][4:6] = (
                left_pix,
                right_pix,
                left_pix,
            )
            out_index += 1
            left_index += 1
            right_index += 1
            continue
        if left_size < right_size:
            # overlapping and left smaller so add left and move left on
            mapping[out_index][0:2], mapping[out_index][2:4], mapping[out_index][4:6] = (
                left_pix,
                right_pix,
                left_pix,
            )
            out_index += 1
            left_index += 1
            continue
        # else overlapping and right smaller so add right and move right on
        mapping[out_index][0:2], mapping[out_index][2:4], mapping[out_index][4:6] = (
            left_pix,
            right_pix,
            right_pix,
        )
        out_index += 1
        right_index += 1
    return mapping[:out_index].T


@njit(
    numba.types.void(
        numba.int64,
        numba.int64,
        numba.int64[:],
        numba.boolean,
        numba.types.List(numba.int64[::1]),
    )
)
def _add_pixels_until(
    add_from: int,
    add_to: int,
    matching_pix: np.ndarray,
    is_left_pixel: bool,
    mapping: List[np.ndarray],
):
    """Adds pixels of the greatest possible order to fill output from `add-from` to `add_to`

    Adds these pixels to the mapping as the aligned pixel, with matching pix added as either the left or right
    side of the mapping and none pixel for the other.

    Args:
        add_from (int): The pixel number to add pixels from
        add_to (int): The pixel number to add pixels to
        matching_pix (int): The matching pixel from the side of the catalog that exists to map the new pixels
            to in the mapping
        is_left_pixel (int): Is the matching pixel from the left side of the alignment
        mapping (List[np.ndarray]): The List mapping left, right, and aligned pixels to add the new pixels to
    """
    while add_from < add_to:
        # maximum power of 4 that is a factor of add_from
        # If add_from is 0, can add any power of 4, so use largest healpix order of 4**29
        max_p4_from = 1 << 58
        if add_from != 0:
            max_p4_from = add_from & -add_from
            if max_p4_from & 0xAAAAAAAAAAAAAAA:
                max_p4_from = max_p4_from >> 1

        # maximum power of 4 less than or equal to (add_to - add_from)
        max_p4_to_log2 = np.int64(np.log2(add_to - add_from))
        max_p4_to = 1 << (max_p4_to_log2 - (max_p4_to_log2 & 1))

        pixel_size = min(max_p4_to, max_p4_from)
        pixel = np.array([add_from, add_from + pixel_size])
        if is_left_pixel:
            mapping.append(np.concatenate((matching_pix, NONE_PIX, pixel)))
        else:
            mapping.append(np.concatenate((NONE_PIX, matching_pix, pixel)))
        add_from = add_from + pixel_size


@njit(
    numba.types.void(
        numba.int64,
        numba.int64[:, :],
        numba.int64,
        numba.boolean,
        numba.types.List(numba.int64[::1]),
    )
)
def _add_remaining_pixels(
    added_until: int,
    pixel_list: np.ndarray,
    index: int,
    is_left_pixel: bool,
    mapping: List[np.ndarray],
):
    """Adds pixels to output and mapping from a given index in a list of pixel intervals

    Args:
        added_until (int): Where the alignment has been added until
        pixel_list (np.ndarray): The list of pixels from the side of the alignment to add the remaining
            pixels of
        index (int): The index of the pixel list to add the pixels from
        is_left_pixel (bool): Is the pixel list from left side of the alignment
        mapping (List[np.ndarray]): The List mapping left, right, and aligned pixels to add the new pixels to
    """

    # first pixel may be partially covered
    pix = pixel_list[index]
    if pix[0] < added_until < pix[1]:
        _add_pixels_until(added_until, pix[1], pix, is_left_pixel, mapping)
        index += 1
    if added_until >= pix[1]:
        index += 1
    while index < len(pixel_list):
        pix = pixel_list[index]
        if is_left_pixel:
            mapping.append(np.concatenate((pix, NONE_PIX, pix)))
        else:
            mapping.append(np.concatenate((NONE_PIX, pix, pix)))
        index += 1


# pylint: disable=too-many-statements
@njit(numba.types.List(numba.int64[::1])(numba.int64[:, :], numba.int64[:, :], numba.boolean, numba.boolean))
def perform_align_trees(
    left: np.ndarray,
    right: np.ndarray,
    include_all_left: bool,
    include_all_right: bool,
) -> List[np.ndarray]:
    """Performs an alignment on arrays of pixel intervals

    Pixel interval lists must be of to the same order

    Args:
        left (np.ndarray): the left array of intervals
        right (np.ndarray): the right array of intervals
        include_all_left (bool): if all pixels from the left tree should be covered in the final alignment
        include_all_right (bool): if all pixels from the right tree should be covered in the final alignment

    Returns (List[np.ndarray]):
        The pixel mapping of the matching left, right, and aligned pixels with each row containing an array of
        [left_order, left_pixel, right_order, right_pixel, aligned_order, aligned_pixel]
    """
    added_until = 0
    mapping = []
    left_index = 0
    right_index = 0
    while left_index < len(left) and right_index < len(right):
        left_pix = left[left_index]
        right_pix = right[right_index]
        if left_pix[0] >= right_pix[1]:
            # left pix ahead of right, no overlap, so move right on
            if include_all_right:
                if added_until <= right_pix[0]:
                    # should cover right pix and no coverage of right pix => add whole right pix
                    mapping.append(np.concatenate((NONE_PIX, right_pix, right_pix)))
                    added_until = right_pix[1]
                elif added_until < right_pix[1]:
                    # should cover right pix and partial coverage of right pix => cover rest of right pix
                    _add_pixels_until(added_until, right_pix[1], right_pix, RIGHT_SIDE, mapping)
                    added_until = right_pix[1]
            right_index += 1
            continue
        if right_pix[0] >= left_pix[1]:
            # right pix ahead of left, no overlap, so move left on
            if include_all_left:
                if added_until <= left_pix[0]:
                    # should cover left pix and no coverage of left pix => add whole left pix
                    mapping.append(np.concatenate((left_pix, NONE_PIX, left_pix)))
                    added_until = left_pix[1]
                elif added_until < left_pix[1]:
                    # should cover left pix and partial coverage of left pix => cover rest of left pix
                    _add_pixels_until(added_until, left_pix[1], left_pix, LEFT_SIDE, mapping)
                    added_until = left_pix[1]
            left_index += 1
            continue
        left_size = left_pix[1] - left_pix[0]
        right_size = right_pix[1] - right_pix[0]
        if left_size == right_size:
            # overlapping & same size => same pixel so add and move both on
            mapping.append(np.concatenate((left_pix, right_pix, left_pix)))
            added_until = left_pix[1]
            left_index += 1
            right_index += 1
            continue
        if left_size < right_size:
            # overlapping and left smaller so add left and move left on
            if include_all_right and left_pix[0] > right_pix[0] and left_pix[0] > added_until:
                # need to cover all of right pix and start of left pix has a gap to current coverage
                # so fill in gap
                add_from = max(added_until, right_pix[0])
                _add_pixels_until(add_from, left_pix[0], right_pix, RIGHT_SIDE, mapping)
            mapping.append(np.concatenate((left_pix, right_pix, left_pix)))
            added_until = left_pix[1]
            left_index += 1
            continue
        # else overlapping and right smaller so add right and move right on

        if include_all_left and right_pix[0] > left_pix[0] and right_pix[0] > added_until:
            # need to cover all of left pix and start of right pix has a gap to current coverage
            # so fill in gap
            add_from = max(added_until, left_pix[0])
            _add_pixels_until(add_from, right_pix[0], left_pix, LEFT_SIDE, mapping)
        mapping.append(np.concatenate((left_pix, right_pix, right_pix)))
        added_until = right_pix[1]
        right_index += 1

    # After loop, if either tree needs to be fully covered and loop hasn't checked all pixels from that tree
    # then cover the remaining pixels
    if include_all_right and right_index < len(right):
        _add_remaining_pixels(added_until, right, right_index, RIGHT_SIDE, mapping)
    if include_all_left and left_index < len(left):
        _add_remaining_pixels(added_until, left, left_index, LEFT_SIDE, mapping)
    return mapping


def filter_alignment_by_moc(alignment: PixelAlignment, moc: MOC) -> PixelAlignment:
    """Filters an alignment by a moc to only include pixels in the aligned_tree that overlap with the moc,
    and the corresponding rows in the mapping.

    Args:
        alignment (PixelAlignment): The pixel alignment to filter
        moc (mocpy.MOC): The moc to filter by

    Returns:
        PixelAlignment object with the filtered mapping and tree
    """
    moc_ranges = moc.to_depth29_ranges
    tree_29_ranges = alignment.pixel_tree.tree << (2 * (29 - alignment.pixel_tree.tree_order))
    tree_mask = perform_filter_by_moc(tree_29_ranges, moc_ranges)
    new_tree = PixelTree(alignment.pixel_tree.tree[tree_mask], alignment.pixel_tree.tree_order)
    return PixelAlignment(new_tree, alignment.pixel_mapping.iloc[tree_mask], alignment.alignment_type)


def align_with_mocs(
    left_tree: PixelTree,
    right_tree: PixelTree,
    left_moc: MOC | None,
    right_moc: MOC | None,
    alignment_type: PixelAlignmentType = PixelAlignmentType.INNER,
) -> PixelAlignment:
    """Aligns two pixel trees and mocs together, resulting in a pixel alignment with only aligned pixels that
    have coverage in the mocs.

    Args:
        left_tree (PixelTree): The left tree to align
        right_tree (PixelTree): The right tree to align
        left_moc (mocpy.MOC): the moc with the coverage of the left catalog
        right_moc (mocpy.MOC): the moc with the coverage of the right catalog
        alignment_type (PixelAlignmentType): The type of alignment describing how to handle nodes which exist
            in one tree but not the other. Options are:

                - inner - only use pixels that appear in both catalogs
                - left - use all pixels that appear in the left catalog and any overlapping from the right
                - right - use all pixels that appear in the right catalog and any overlapping from the left
                - outer - use all pixels from both catalogs

    Returns:
        The PixelAlignment object with the aligned trees filtered by the coverage in the catalogs.
    """
    left_moc = left_moc if left_moc is not None else left_tree.to_moc()
    right_moc = right_moc if right_moc is not None else right_tree.to_moc()
    moc_intersection_methods: Dict[PixelAlignmentType, Callable[[MOC, MOC], MOC]] = {
        PixelAlignmentType.INNER: lambda l, r: l.intersection(r),
        PixelAlignmentType.LEFT: lambda l, r: l,
        PixelAlignmentType.RIGHT: lambda l, r: r,
        PixelAlignmentType.OUTER: lambda l, r: l.union(r),
    }
    filter_moc = moc_intersection_methods[alignment_type](left_moc, right_moc)
    alignment = align_trees(left_tree, right_tree, alignment_type=alignment_type)
    return filter_alignment_by_moc(alignment, filter_moc)



"""Test catalog information is valid"""

import os
import shutil
from pathlib import Path

import pytest

from hats.catalog import PartitionInfo
from hats.io.validation import is_valid_catalog


def test_is_valid_catalog(tmp_path, small_sky_catalog, small_sky_pixels):
    """Tests if the catalog_info and partition_info files are valid"""
    # An empty directory means an invalid catalog
    assert not is_valid_catalog(tmp_path)

    # Having the catalog_info file is not enough
    small_sky_catalog.catalog_info.to_properties_file(tmp_path)
    assert not is_valid_catalog(tmp_path)

    # The catalog is valid if both the catalog_info and _metadata files exist,
    # and the catalog_info is in a valid format
    total_rows = PartitionInfo.from_healpix(small_sky_pixels).write_to_metadata_files(tmp_path)
    assert total_rows == 1
    assert is_valid_catalog(tmp_path)

    # A partition_info file alone is also not enough
    os.remove(tmp_path / "properties")
    assert not is_valid_catalog(tmp_path)

    # The catalog_info file needs to be in the correct format
    small_sky_catalog.catalog_info.catalog_type = "invalid"
    small_sky_catalog.catalog_info.to_properties_file(tmp_path)
    assert not is_valid_catalog(tmp_path)


def test_is_valid_catalog_strict(tmp_path, small_sky_catalog, small_sky_pixels, small_sky_order1_pixels):
    """Tests if the catalog_info and partition_info files are valid"""
    # Set up our arguments once, just to be sure we're calling validation
    # the same way throughout this test scene.
    flags = {
        "strict": True,  # more intensive checks
        "fail_fast": False,  # check everything, and just return true/false
        "verbose": False,  # don't print, throw warnings.
    }
    # An empty directory means an invalid catalog
    with pytest.warns():
        assert not is_valid_catalog(tmp_path, **flags)

    # Having the catalog_info file is not enough
    small_sky_catalog.catalog_info.to_properties_file(tmp_path)
    with pytest.warns():
        assert not is_valid_catalog(tmp_path, **flags)

    # Adds the _metadata and _common_metadata, but that's not enough.
    total_rows = PartitionInfo.from_healpix(small_sky_pixels).write_to_metadata_files(tmp_path)
    assert total_rows == 1
    with pytest.warns():
        assert not is_valid_catalog(tmp_path, **flags)

    # Adds the partition_info.csv, but that's STILL not enough
    PartitionInfo.from_healpix(small_sky_pixels).write_to_file(catalog_path=tmp_path)
    with pytest.warns():
        assert not is_valid_catalog(tmp_path, **flags)

    # This outta do it! Add parquet files that match the _metadata pixels.
    shutil.copytree(
        Path(small_sky_catalog.catalog_path) / "dataset" / "Norder=0",
        tmp_path / "dataset" / "Norder=0",
    )

    assert is_valid_catalog(tmp_path, **flags)

    # Uh oh! Now the sets of pixels don't match between _metadata and partition_info.csv!
    PartitionInfo.from_healpix(small_sky_order1_pixels).write_to_file(catalog_path=tmp_path)
    with pytest.warns():
        assert not is_valid_catalog(tmp_path, **flags)


def test_is_valid_catalog_fail_fast(tmp_path, small_sky_catalog, small_sky_pixels):
    """Tests if the catalog_info and partition_info files are valid"""
    # Set up our arguments once, just to be sure we're calling validation
    # the same way throughout this test scene.
    flags = {
        "strict": True,  # more intensive checks
        "fail_fast": True,  # raise an error at the first problem.
        "verbose": False,  # don't bother printing anything.
    }
    # An empty directory means an invalid catalog
    with pytest.raises(ValueError, match="properties file"):
        is_valid_catalog(tmp_path, **flags)

    # Having the catalog_info file is not enough
    small_sky_catalog.catalog_info.to_properties_file(tmp_path)
    with pytest.raises(ValueError, match="partition_info.csv"):
        is_valid_catalog(tmp_path, **flags)

    PartitionInfo.from_healpix(small_sky_pixels).write_to_file(catalog_path=tmp_path)
    with pytest.raises(ValueError, match="_metadata"):
        is_valid_catalog(tmp_path, **flags)

    total_rows = PartitionInfo.from_healpix(small_sky_pixels).write_to_metadata_files(tmp_path)
    assert total_rows == 1
    with pytest.raises(ValueError, match="parquet paths"):
        is_valid_catalog(tmp_path, **flags)

    shutil.copytree(
        Path(small_sky_catalog.catalog_path) / "dataset" / "Norder=0",
        tmp_path / "dataset" / "Norder=0",
    )
    assert is_valid_catalog(tmp_path, **flags)


def test_is_valid_catalog_verbose_fail(tmp_path, capsys):
    """Tests if the catalog_info and partition_info files are valid"""
    # Set up our arguments once, just to be sure we're calling validation
    # the same way throughout this test scene.
    flags = {
        "strict": True,  # more intensive checks
        "fail_fast": False,  # check everything, and just return true/false
        "verbose": True,  # print messages along the way
    }
    # An empty directory means an invalid catalog
    assert not is_valid_catalog(tmp_path, **flags)

    captured = capsys.readouterr().out
    assert "Validating catalog at path" in captured
    assert "properties file does not exist or is invalid" in captured
    assert "partition_info.csv file does not exist" in captured
    assert "_metadata file does not exist" in captured
    assert "_common_metadata file does not exist" in captured


def test_is_valid_catalog_verbose_success(small_sky_dir, capsys):
    """Tests if the catalog_info and partition_info files are valid"""
    # Set up our arguments once, just to be sure we're calling validation
    # the same way throughout this test scene.
    flags = {
        "strict": True,  # more intensive checks
        "fail_fast": False,  # check everything, and just return true/false
        "verbose": True,  # print messages along the way
    }
    assert is_valid_catalog(small_sky_dir, **flags)
    captured = capsys.readouterr().out
    assert "Validating catalog at path" in captured
    assert "Found 1 partition" in captured
    assert "Approximate coverage is 3437.75 sq deg" in captured


def test_valid_catalog_strict_all(small_sky_source_dir, small_sky_order1_dir, small_sky_dir):
    """Check that all of our object catalogs in test data are valid, using strict mechanism"""
    flags = {
        "strict": True,  # more intensive checks
        "fail_fast": False,  # check everything, and just return true/false
        "verbose": False,  # don't bother printing anything.
    }
    assert is_valid_catalog(small_sky_source_dir, **flags)
    assert is_valid_catalog(small_sky_order1_dir, **flags)
    assert is_valid_catalog(small_sky_dir, **flags)



"""Utilities for generating and manipulating object count histograms"""

import numpy as np
import pandas as pd

import hats.pixel_math.healpix_shim as hp


def empty_histogram(highest_order):
    """Use numpy to create an histogram array with the right shape, filled with zeros.

    Args:
        highest_order (int): the highest healpix order (e.g. 0-10)
    Returns:
        one-dimensional numpy array of long integers, where the length is equal to
        the number of pixels in a healpix map of target order, and all values are set to 0.
    """
    return np.zeros(hp.order2npix(highest_order), dtype=np.int64)


def generate_histogram(
    data: pd.DataFrame,
    highest_order,
    ra_column="ra",
    dec_column="dec",
):
    """Generate a histogram of counts for objects found in `data`

    Args:
        data (:obj:`pd.DataFrame`): tabular object data
        highest_order (int):  the highest healpix order (e.g. 0-10)
        ra_column (str): where in the input to find the celestial coordinate, right ascension
        dec_column (str): where in the input to find the celestial coordinate, declination
    Returns:
        one-dimensional numpy array of long integers where the value at each index corresponds
        to the number of objects found at the healpix pixel.
    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
    """
    histogram_result = empty_histogram(highest_order)

    # Verify that the data frame has columns with desired names.
    required_columns = [ra_column, dec_column]
    if not all(x in data.columns for x in required_columns):
        raise ValueError(f"Invalid column names in input: {ra_column}, {dec_column}")
    mapped_pixels = hp.ang2pix(
        2**highest_order,
        data[ra_column].values,
        data[dec_column].values,
        lonlat=True,
        nest=True,
    )
    mapped_pixel, count_at_pixel = np.unique(mapped_pixels, return_counts=True)
    histogram_result[mapped_pixel] += count_at_pixel.astype(np.int64)
    return histogram_result


def generate_alignment(
    histogram, highest_order=10, lowest_order=0, threshold=1_000_000, drop_empty_siblings=False
):
    """Generate alignment from high order pixels to those of equal or lower order

    We may initially find healpix pixels at order 10, but after aggregating up to the pixel
    threshold, some final pixels are order 4 or 7. This method provides a map from pixels
    at order 10 to their destination pixel. This may be used as an input into later partitioning
    map reduce steps.

    Args:
        histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        highest_order (int):  the highest healpix order (e.g. 5-10)
        lowest_order (int): the lowest healpix order (e.g. 1-5). specifying a lowest order
            constrains the partitioning to prevent spatially large pixels.
        threshold (int): the maximum number of objects allowed in a single pixel
        drop_empty_siblings (bool): if 3 of 4 pixels are empty, keep only the non-empty pixel
    Returns:
        one-dimensional numpy array of integer 3-tuples, where the value at each index corresponds
        to the destination pixel at order less than or equal to the `highest_order`.

        The tuple contains three integers:

        - order of the destination pixel
        - pixel number *at the above order*
        - the number of objects in the pixel
    Raises:
        ValueError: if the histogram is the wrong size, or some initial histogram bins
            exceed threshold.
    """
    if len(histogram) != hp.order2npix(highest_order):
        raise ValueError("histogram is not the right size")
    if lowest_order > highest_order:
        raise ValueError("lowest_order should be less than highest_order")
    max_bin = np.amax(histogram)
    if max_bin > threshold:
        raise ValueError(f"single pixel count {max_bin} exceeds threshold {threshold}")

    nested_sums = []
    for i in range(0, highest_order):
        nested_sums.append(empty_histogram(i))
    nested_sums.append(histogram)

    # work backward - from highest order, fill in the sums of lower order pixels
    for read_order in range(highest_order, lowest_order, -1):
        parent_order = read_order - 1
        for index in range(0, len(nested_sums[read_order])):
            parent_pixel = index >> 2
            nested_sums[parent_order][parent_pixel] += nested_sums[read_order][index]

    if drop_empty_siblings:
        return _get_alignment_dropping_siblings(nested_sums, highest_order, lowest_order, threshold)
    return _get_alignment(nested_sums, highest_order, lowest_order, threshold)


def _get_alignment(nested_sums, highest_order, lowest_order, threshold):
    """Method to aggregate pixels up to the threshold.

    Checks from low order (large areas), drilling down into higher orders (smaller areas) to
    find the appropriate order for an area of sky."""
    nested_alignment = []
    for i in range(0, highest_order + 1):
        nested_alignment.append(np.full(hp.order2npix(i), None))

    # work forward - determine if we should map to a lower order pixel, this pixel, or keep looking.
    for read_order in range(lowest_order, highest_order + 1):
        parent_order = read_order - 1
        for index in range(0, len(nested_sums[read_order])):
            parent_alignment = None
            if parent_order >= 0:
                parent_pixel = index >> 2
                parent_alignment = nested_alignment[parent_order][parent_pixel]

            if parent_alignment:
                nested_alignment[read_order][index] = parent_alignment
            elif nested_sums[read_order][index] == 0:
                continue
            elif nested_sums[read_order][index] <= threshold:
                nested_alignment[read_order][index] = (
                    read_order,
                    index,
                    nested_sums[read_order][index],
                )

    return nested_alignment[highest_order]


def _get_alignment_dropping_siblings(nested_sums, highest_order, lowest_order, threshold):
    """Method to aggregate pixels up to the threshold that collapses completely empty pixels away.

    Checks from higher order (smaller areas) out to lower order (large areas). In this way, we are able to
    keep spatially isolated areas in pixels of higher order.

    This method can be slower than the above `_get_alignment` method, and so should only be used
    when the smaller area pixels are desired.

    This uses a form of hiearchical agglomeration (building a tree bottom-up). For each cell
    at order n, we look at the counts in all 4 subcells at order (n+1). We have two numeric
    values that are easy to compute that we can refer to easily:

    - quad_sum: the total number of counts in this cell
    - quad_max: the largest count within the 4 subcells

    Our agglomeration criteria (the conditions under which we collapse) must BOTH be met:

    - total number in cell is less than the global threshold (quad_sum <= threshold)
    - more than one subcell contains values (quad_sum != quad_max) (if exactly 1
      subcell contains counts, then all of the quad_sum will come from that single quad_max)

    Inversely, we will NOT collapse when EITHER is true:

    - total number in cell is greater than the threshold
    - only one subcell contains values
    """
    order_map = np.array(
        [highest_order if count > 0 else -1 for count in nested_sums[highest_order]], dtype=np.int32
    )
    for pixel_order in range(highest_order - 1, lowest_order - 1, -1):
        for quad_start_index in range(0, hp.order2npix(pixel_order)):
            quad_sum = nested_sums[pixel_order][quad_start_index]
            quad_max = max(nested_sums[pixel_order + 1][quad_start_index * 4 : quad_start_index * 4 + 4])

            if quad_sum != quad_max and quad_sum <= threshold:
                ## Condition where we want to collapse pixels to the lower order (larger area)
                explosion_factor = 4 ** (highest_order - pixel_order)
                exploded_pixels = [
                    *range(
                        quad_start_index * explosion_factor,
                        (quad_start_index + 1) * explosion_factor,
                    )
                ]
                order_map[exploded_pixels] = pixel_order

    # Construct our results.
    nested_alignment = [
        (
            (intended_order, pixel_high_index >> 2 * (highest_order - intended_order))
            if intended_order >= 0
            else None
        )
        for pixel_high_index, intended_order in enumerate(order_map)
    ]
    nested_alignment = [
        (tup[0], tup[1], nested_sums[tup[0]][tup[1]]) if tup else None for tup in nested_alignment
    ]

    return np.array(nested_alignment, dtype="object")



import pandas as pd
import pytest

from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.pixel_math.healpix_pixel import HealpixPixel


def test_init(association_catalog_join_pixels):
    partition_join_info = PartitionJoinInfo(association_catalog_join_pixels)
    pd.testing.assert_frame_equal(partition_join_info.data_frame, association_catalog_join_pixels)


def test_wrong_columns(association_catalog_join_pixels):
    for column in PartitionJoinInfo.COLUMN_NAMES:
        join_pixels = association_catalog_join_pixels.copy()
        join_pixels = join_pixels.rename(columns={column: "wrong_name"})
        with pytest.raises(ValueError, match=column):
            PartitionJoinInfo(join_pixels)


def test_read_from_metadata(association_catalog_join_pixels, association_catalog_path):
    info = PartitionJoinInfo.read_from_file(association_catalog_path / "dataset" / "_metadata")
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)


def test_read_from_metadata_fail(tmp_path):
    empty_dataframe = pd.DataFrame()
    metadata_filename = tmp_path / "empty_metadata.parquet"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.raises(ValueError, match="missing columns"):
        PartitionJoinInfo.read_from_file(metadata_filename)

    with pytest.raises(ValueError, match="at least one column"):
        PartitionJoinInfo.read_from_file(metadata_filename, strict=True)

    ## Starting with a valid join info, remove each column and make sure we error.
    valid_ish_dataframe = pd.DataFrame(
        {"data": [0], "Norder": [3], "Npix": [45], "join_Norder": [3], "join_Npix": [45]}
    )
    metadata_filename = tmp_path / "test_metadata.parquet"
    valid_ish_dataframe.to_parquet(metadata_filename)
    PartitionJoinInfo.read_from_file(metadata_filename)

    for drop_column in ["Norder", "Npix", "join_Norder", "join_Npix"]:
        missing_column_dataframe = valid_ish_dataframe.drop(labels=drop_column, axis=1)
        missing_column_dataframe.to_parquet(metadata_filename)
        with pytest.raises(ValueError, match=f"missing .*{drop_column}"):
            PartitionJoinInfo.read_from_file(metadata_filename)

    with pytest.raises(ValueError, match="empty file path"):
        PartitionJoinInfo.read_from_file(metadata_filename, strict=True)


def test_load_partition_join_info_from_dir_fail(tmp_path):
    empty_dataframe = pd.DataFrame()
    metadata_filename = tmp_path / "empty_metadata.parquet"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.raises(FileNotFoundError, match="_metadata or partition join info"):
        PartitionJoinInfo.read_from_dir(tmp_path)

    # The file is there, but doesn't have the required content.
    (tmp_path / "dataset").mkdir()
    metadata_filename = tmp_path / "dataset" / "_metadata"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.warns(UserWarning, match="slow"):
        with pytest.raises(ValueError, match="missing columns"):
            PartitionJoinInfo.read_from_dir(tmp_path)


def test_primary_to_join_map(association_catalog_join_pixels):
    info = PartitionJoinInfo(association_catalog_join_pixels)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)
    pixel_map = info.primary_to_join_map()

    expected = {
        HealpixPixel(0, 11): [
            HealpixPixel(1, 44),
            HealpixPixel(1, 45),
            HealpixPixel(1, 46),
            HealpixPixel(1, 47),
        ]
    }
    assert pixel_map == expected


def test_metadata_file_round_trip(association_catalog_join_pixels, tmp_path):
    info = PartitionJoinInfo(association_catalog_join_pixels)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)
    total_rows = info.write_to_metadata_files(tmp_path)
    assert total_rows == 4

    new_info = PartitionJoinInfo.read_from_file(tmp_path / "dataset" / "_metadata")
    pd.testing.assert_frame_equal(new_info.data_frame, association_catalog_join_pixels)


def test_csv_file_round_trip(association_catalog_join_pixels, tmp_path):
    info = PartitionJoinInfo(association_catalog_join_pixels)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)
    info.write_to_csv(tmp_path)

    new_info = PartitionJoinInfo.read_from_csv(tmp_path / "partition_join_info.csv")
    pd.testing.assert_frame_equal(new_info.data_frame, association_catalog_join_pixels)


def test_read_from_csv(association_catalog_partition_join_file, association_catalog_join_pixels):
    info = PartitionJoinInfo.read_from_csv(association_catalog_partition_join_file)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)


def test_read_from_missing_file(tmp_path):
    with pytest.raises(FileNotFoundError):
        PartitionJoinInfo.read_from_csv(tmp_path / "wrong")


def test_load_partition_info_from_dir_and_write(tmp_path, association_catalog_join_pixels):
    info = PartitionJoinInfo(association_catalog_join_pixels)

    ## Path arguments are required if the info was not created from a `read_from_dir` call
    with pytest.raises(ValueError):
        info.write_to_csv()
    with pytest.raises(ValueError):
        info.write_to_metadata_files()

    info.write_to_csv(catalog_path=tmp_path)
    info = PartitionJoinInfo.read_from_dir(tmp_path)

    ## Can write out the partition info CSV by providing:
    ##  - no arguments
    ##  - new catalog directory
    info.write_to_csv()
    info.write_to_csv(catalog_path=tmp_path)

    ## Can write out the _metadata file by providing:
    ##  - no arguments
    ##  - new catalog directory
    total_rows = info.write_to_metadata_files()
    assert total_rows == 4

    total_rows = info.write_to_metadata_files(catalog_path=tmp_path)
    assert total_rows == 4



from typing import List

from hats.pixel_math import HealpixPixel
from hats.pixel_tree import PixelAlignmentType, align_trees
from hats.pixel_tree.pixel_tree import PixelTree


def compute_negative_tree_pixels(tree: PixelTree) -> List[HealpixPixel]:
    """Computes a 'negative pixel tree' consisting of the pixels needed to cover the full sky not in the tree

    Args:
        tree (PixelTree): the input tree to compute the negative of

    Returns (List[HealpixPixel]):
        A list of HEALPix pixels needed to cover the part of the sky not covered by the tree, using the least
        number of pixels possible.
    """
    tree_pixels = set(tree.get_healpix_pixels())
    full_tree = PixelTree.from_healpix([(0, i) for i in range(12)])
    alignment = align_trees(full_tree, tree, alignment_type=PixelAlignmentType.OUTER)
    aligned_tree = alignment.pixel_tree
    return [p for p in aligned_tree.get_healpix_pixels() if p not in tree_pixels]



from __future__ import annotations

from typing import List

import numpy as np

import hats.pixel_math.healpix_shim as hp

SPATIAL_INDEX_COLUMN = "_healpix_29"
SPATIAL_INDEX_ORDER = 29


def compute_spatial_index(ra_values: List[float], dec_values: List[float]) -> np.ndarray:
    """Compute the healpix index field.

    Args:
        ra_values (List[float]): celestial coordinates, right ascension in degrees
        dec_values (List[float]): celestial coordinates, declination in degrees
    Returns:
        one-dimensional numpy array of int64s with healpix NESTED pixel numbers at order 29
    Raises:
        ValueError: if the length of the input lists don't match.
    """
    if len(ra_values) != len(dec_values):
        raise ValueError("ra and dec arrays should have the same length")

    mapped_pixels = hp.ang2pix(2**SPATIAL_INDEX_ORDER, ra_values, dec_values, nest=True, lonlat=True)

    return mapped_pixels


def spatial_index_to_healpix(ids: List[int], target_order: int = SPATIAL_INDEX_ORDER) -> np.ndarray:
    """Convert _healpix_29 values to the healpix pixel at the specified order

    Args:
        ids (List[int64]): list of well-formatted _healpix_29 values
        target_order (int64): Defaults to `SPATIAL_INDEX_ORDER`.
            The order of the pixel to get from the healpix index.
    Returns:
        numpy array of target_order pixels from the healpix index
    """
    delta_order = SPATIAL_INDEX_ORDER - target_order
    return np.array(ids) >> (2 * delta_order)


def healpix_to_spatial_index(order: int | List[int], pixel: int | List[int]) -> np.int64 | np.ndarray:
    """Convert a healpix pixel to the healpix index

    This maps the healpix pixel to the lowest pixel number within that pixel at order 29.

    Useful for operations such as filtering by _healpix_29.

    Args:
        order (int64 | List[int64]): order of pixel to convert
        pixel (int64 | List[int64]): pixel number in nested ordering of pixel to convert
    Returns:
        healpix index or numpy array of healpix indices
    """
    order = np.int64(order)
    pixel = np.int64(pixel)
    pixel_higher_order = pixel * (4 ** (SPATIAL_INDEX_ORDER - order))
    return pixel_higher_order



from __future__ import annotations

from dataclasses import dataclass
from typing import List

import numpy as np

from hats.pixel_math.spatial_index import SPATIAL_INDEX_ORDER


@dataclass(eq=True, frozen=True, order=True)
class HealpixPixel:
    """A HEALPix pixel, represented by an order and pixel number in NESTED ordering scheme

    see https://lambda.gsfc.nasa.gov/toolbox/pixelcoords.html for more information
    """

    order: int
    pixel: int

    def __post_init__(self) -> None:
        """Initialize a HEALPix pixel
        Args:
            order: HEALPix order
            pixel: HEALPix pixel number in NESTED ordering scheme
        """
        if self.order > SPATIAL_INDEX_ORDER:
            raise ValueError(f"HEALPix order cannot be greater than {SPATIAL_INDEX_ORDER}")

    def __str__(self) -> str:
        return f"Order: {self.order}, Pixel: {self.pixel}"

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, key: int) -> int:
        if key == 0:
            return self.order
        if key == 1:
            return self.pixel
        raise IndexError("Invalid healpix index")

    def convert_to_lower_order(self, delta_order: int) -> HealpixPixel:
        """Returns the HEALPix pixel that contains the pixel at a lower order

        Args:
            delta_order: the difference in order to be subtracted from the current order to generate
                the pixel at a lower order. Must be non-negative

        Returns:
            A new `HealpixPixel` at the current order - `delta_order` which contains the current
            pixel

        Raises:
            ValueError: If delta_order is greater than the current order, a pixel cannot be
                generated at a negative order. Or if delta_order is negative

        """
        new_pixel = get_lower_order_pixel(self.order, self.pixel, delta_order)
        new_order = self.order - delta_order
        return HealpixPixel(new_order, new_pixel)

    def convert_to_higher_order(self, delta_order: int) -> List[HealpixPixel]:
        """Returns a list of HEALPix pixels making up the current pixel at a higher order

        Args:
            delta_order: the difference in order to be added to the current order to generate the
                pixels at a higher order. Must be non-negative

        Returns:
            A new `HealpixPixel` at the current order - `delta_order` which contains the current
            pixel

        Raises:
            ValueError: If delta_order + current order is greater than the maximum HEALPix order,
                pixels cannot be  generated. Or if delta_order is negative
        """
        new_pixels = get_higher_order_pixels(self.order, self.pixel, delta_order)
        new_order = self.order + delta_order
        pixels = [HealpixPixel(new_order, new_pixel) for new_pixel in new_pixels]
        return pixels

    @property
    def dir(self) -> int:
        """Directory number for the pixel.

        This is necessary for file systems that limit to 10,000 subdirectories.
        The directory name will take the HiPS standard form of::

            <catalog_base_dir>/Norder=<pixel_order>/Dir=<directory number>

        Where the directory number is calculated using integer division as::

            (pixel_number/10000)*10000
        """
        return int(self.pixel / 10_000) * 10_000


INVALID_PIXEL = HealpixPixel(-1, -1)


def get_lower_order_pixel(order: int, pixel: int, delta_order: int) -> int:
    """Returns the pixel number at a lower order

    Args:
        order (int): the order of the pixel
        pixel (int): the pixel number of the pixel in NESTED ordering
        delta_order (int): the change in order to the new lower order

    Returns:
        The pixel number at order (order - delta_order) for the pixel that contains the given pixel
    """
    if order - delta_order < 0:
        raise ValueError("Pixel Order cannot be below zero")
    if delta_order < 0:
        raise ValueError("delta order cannot be below zero")
    new_pixel = pixel >> (2 * delta_order)
    return new_pixel


def get_higher_order_pixels(order: int, pixel: int, delta_order: int) -> np.ndarray:
    """Returns the pixel numbers at a higher order

    Args:
        order (int): the order of the pixel
        pixel (int): the pixel number of the pixel in NESTED ordering
        delta_order (int): the change in order to the new higher order

    Returns:
        The list of pixel numbers at order (order + delta_order) for the pixels contained by the given pixel
    """
    if order == -1:
        return np.arange(0, 12)
    if order + delta_order > SPATIAL_INDEX_ORDER:
        raise ValueError(f"Pixel Order cannot be above maximum order {SPATIAL_INDEX_ORDER}")
    if delta_order < 0:
        raise ValueError("delta order cannot be below zero")
    new_pixels = np.arange(pixel << (2 * delta_order), (pixel + 1) << (2 * delta_order))
    return new_pixels



from enum import Enum


class PixelAlignmentType(str, Enum):
    """The type of alignment between catalog trees. Like a join, the types describe if nodes should
    be included if they do not appear in the other catalog"""

    INNER = "inner"
    OUTER = "outer"
    LEFT = "left"
    RIGHT = "right"



from .file_io import (
    delete_file,
    load_csv_to_pandas,
    load_csv_to_pandas_generator,
    load_text_file,
    make_directory,
    read_fits_image,
    read_parquet_dataset,
    read_parquet_file_to_pandas,
    read_parquet_metadata,
    remove_directory,
    write_dataframe_to_csv,
    write_dataframe_to_parquet,
    write_fits_image,
    write_parquet_metadata,
    write_string_to_file,
)
from .file_pointer import (
    append_paths_to_pointer,
    directory_has_contents,
    does_file_or_directory_exist,
    find_files_matching_path,
    get_directory_contents,
    get_upath,
    is_regular_file,
)



import numpy.testing as npt

from hats.catalog.index.index_catalog import IndexCatalog
from hats.loaders import read_hats
from hats.pixel_math import HealpixPixel


def test_loc_partition(small_sky_source_object_index_dir):
    catalog = read_hats(small_sky_source_object_index_dir)

    assert isinstance(catalog, IndexCatalog)
    assert catalog.on_disk
    assert catalog.catalog_path == small_sky_source_object_index_dir

    npt.assert_array_equal(catalog.loc_partitions([700]), [HealpixPixel(2, 184)])
    npt.assert_array_equal(catalog.loc_partitions([707]), [HealpixPixel(2, 176), HealpixPixel(2, 178)])
    npt.assert_array_equal(catalog.loc_partitions([900]), [])



"""Utilities for find the pixels of higher orders that surround a given healpixel."""

import numpy as np

import hats.pixel_math.healpix_shim as hp

# see the documentation for get_edge() about this variable
_edge_vectors = [
    np.asarray([1, 3]),  # NE edge
    np.asarray([1]),  # E corner
    np.asarray([0, 1]),  # SE edge
    np.asarray([0]),  # S corner
    np.asarray([0, 2]),  # SW edge
    np.asarray([2]),  # W corner
    np.asarray([2, 3]),  # NW edge
    np.asarray([3]),  # N corner
]

# cache used by get_margin()
_suffixes = {}


def get_edge(d_order, pix, edge):
    """Get all the pixels at order kk+dk bordering pixel pix.
    See hats/pixel_math/README.md for more info.

    Args:
        dk (int): the change in k that we wish to find the margins for.
        pix (int): the healpix pixel to find margin pixels of.
        edge (int): the edge we want to find for the given pixel. (0-7)

            - 0: NE edge
            - 1: E corner
            - 2: SE edge
            - 3: S corner
            - 4: SW edge
            - 5: W corner
            - 6: NW edge
            - 7: N corner
    Returns:
        one-dimensional numpy array of long integers, filled with the healpix pixels
        at order kk+dk that border the given edge of pix.
    """
    if edge < 0 or edge > 7:
        raise ValueError("edge can only be values between 0 and 7 (see docstring)")

    if (edge, d_order) in _suffixes:
        pixel_edge = _suffixes[(edge, d_order)]
    else:
        # generate and cache the suffix:

        # generate all combinations of i,j,k,... suffixes for the requested edge
        # See https://stackoverflow.com/a/35608701
        grid = np.array(np.meshgrid(*[_edge_vectors[edge]] * d_order))
        pixel_edge = grid.T.reshape(-1, d_order)
        # bit-shift each suffix by the required number of bits
        pixel_edge <<= np.arange(2 * (d_order - 1), -2, -2)
        # sum them up row-wise, generating the suffix
        pixel_edge = pixel_edge.sum(axis=1)
        # cache for further reuse
        _suffixes[(edge, d_order)] = pixel_edge

    # append the 'pix' preffix
    pixel_edge = (pix << 2 * d_order) + pixel_edge
    return pixel_edge


def get_margin(order, pix, d_order):
    """Get all the pixels at order order+dk bordering pixel pix.
    See hats/pixel_math/README.md for more info.

    Args:
        order (int): the healpix order of pix.
        pix (int): the healpix pixel to find margin pixels of.
        d_order (int): the change in k that we wish to find the margins for.
            Must be greater than kk.
    Returns:
        one-dimensional numpy array of long integers, filled with the healpix pixels
        at order kk+dk that border pix.
    """
    if d_order < 1:
        raise ValueError("dk must be greater than order")

    nside = hp.order2nside(order)

    # get all neighboring pixels
    neighbors = hp.get_all_neighbours(nside, pix, nest=True)

    # get the healpix faces IDs of pix and the neighboring pixels
    _, _, pix_face = hp.pix2xyf(nside, pix, nest=True)
    _, _, faces = hp.pix2xyf(nside, neighbors, nest=True)

    # indices which tell get_edge() which edge/vertex to return
    # for a given pixel. The default order is compatible with the
    # order returned by hp.get_all_neighbours().
    which = np.arange(8)

    # northern hemisphere; 90deg cw rotation for every +1 face increment
    if pix_face < 4:
        mask = faces < 4
        which[mask] += 2 * (faces - pix_face)[mask]
        which %= 8
    # southern hemisphere; 90deg ccw rotation for every +1 face increment
    elif pix_face >= 8:
        mask = faces >= 8
        which[mask] -= 2 * (faces - pix_face)[mask]
        which %= 8

    # get all edges/vertices
    # (making sure we skip -1 entries, for pixels with seven neighbors)
    margins = []
    for edge, pixel in zip(which, neighbors):
        if pixel != -1:
            margins.append(get_edge(d_order, pixel, edge))
    margins = np.concatenate(margins)
    return margins



from __future__ import annotations

from typing import Tuple

import numpy as np
from mocpy import MOC
from typing_extensions import TypeAlias

import hats.pixel_math.healpix_shim as hp

# Pair of spherical sky coordinates (ra, dec)
SphericalCoordinates: TypeAlias = Tuple[float, float]

# Sky coordinates on the unit sphere, in cartesian representation (x,y,z)
CartesianCoordinates: TypeAlias = Tuple[float, float, float]


def generate_polygon_moc(vertices: np.array, order: int) -> MOC:
    """Generates a moc filled with leaf nodes at a given order that overlap within
    a polygon. Vertices is an array of cartesian coordinates, in representation (x,y,z)
    and shape (Num vertices, 3), representing the vertices of the polygon."""
    polygon_pixels = hp.query_polygon(hp.order2nside(order), vertices, inclusive=True, nest=True)
    polygon_orders = np.full(len(polygon_pixels), fill_value=order)
    return MOC.from_healpix_cells(ipix=polygon_pixels, depth=polygon_orders, max_depth=order)



import numpy as np
from mocpy import MOC

from hats.pixel_math import HealpixPixel
from hats.pixel_tree.moc_filter import filter_by_moc


def test_moc_filter(pixel_tree_2):
    orders = np.array([1, 1, 2])
    pixels = np.array([45, 46, 128])
    moc = MOC.from_healpix_cells(pixels, orders, 2)
    filtered_tree = filter_by_moc(pixel_tree_2, moc)
    assert filtered_tree.get_healpix_pixels() == [
        HealpixPixel(2, 128),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
    ]


def test_moc_filter_lower_order(pixel_tree_2):
    orders = np.array([0, 1])
    pixels = np.array([11, 32])
    moc = MOC.from_healpix_cells(pixels, orders, 1)
    filtered_tree = filter_by_moc(pixel_tree_2, moc)
    assert filtered_tree.get_healpix_pixels() == [
        HealpixPixel(2, 128),
        HealpixPixel(2, 130),
        HealpixPixel(2, 131),
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
    ]


def test_moc_filter_higher_order(pixel_tree_2):
    orders = np.array([1, 3])
    pixels = np.array([40, 520])
    moc = MOC.from_healpix_cells(pixels, orders, 3)
    filtered_tree = filter_by_moc(pixel_tree_2, moc)
    assert filtered_tree.get_healpix_pixels() == [
        HealpixPixel(2, 130),
        HealpixPixel(0, 10),
    ]


def test_moc_filter_empty_moc(pixel_tree_2):
    orders = np.array([])
    pixels = np.array([])
    moc = MOC.from_healpix_cells(pixels, orders, 0)
    filtered_tree = filter_by_moc(pixel_tree_2, moc)
    assert filtered_tree.get_healpix_pixels() == []


def test_moc_filter_empty_result(pixel_tree_2):
    orders = np.array([0])
    pixels = np.array([1])
    moc = MOC.from_healpix_cells(pixels, orders, 0)
    filtered_tree = filter_by_moc(pixel_tree_2, moc)
    assert filtered_tree.get_healpix_pixels() == []



"""Container class to hold per-partition metadata"""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import pyarrow as pa
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.io import file_io, paths
from hats.io.parquet_metadata import (
    read_row_group_fragments,
    row_group_stat_single_value,
    write_parquet_metadata_for_batches,
)
from hats.pixel_math import HealpixPixel


class PartitionInfo:
    """Container class for per-partition info."""

    METADATA_ORDER_COLUMN_NAME = "Norder"
    METADATA_PIXEL_COLUMN_NAME = "Npix"

    def __init__(self, pixel_list: List[HealpixPixel], catalog_base_dir: str = None) -> None:
        self.pixel_list = pixel_list
        self.catalog_base_dir = catalog_base_dir

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Get healpix pixel objects for all pixels represented as partitions.

        Returns:
            List of HealpixPixel
        """
        return self.pixel_list

    def get_highest_order(self) -> int:
        """Get the highest healpix order for the dataset.

        Returns:
            int representing highest order.
        """
        max_pixel = np.max(self.pixel_list)
        return max_pixel.order

    def write_to_file(
        self,
        partition_info_file: str | Path | UPath | None = None,
        catalog_path: str | Path | UPath | None = None,
    ):
        """Write all partition data to CSV file.

        If no paths are provided, the catalog base directory from the `read_from_dir` call is used.

        Args:
            partition_info_file: path to where the `partition_info.csv`
                file will be written.
            catalog_path: base directory for a catalog where the `partition_info.csv`
                file will be written.

        Raises:
            ValueError: if no path is provided, and could not be inferred.
        """
        if partition_info_file is None:
            if catalog_path is not None:
                partition_info_file = paths.get_partition_info_pointer(catalog_path)
            elif self.catalog_base_dir is not None:
                partition_info_file = paths.get_partition_info_pointer(self.catalog_base_dir)
            else:
                raise ValueError("partition_info_file is required if info was not loaded from a directory")

        file_io.write_dataframe_to_csv(self.as_dataframe(), partition_info_file, index=False)

    def write_to_metadata_files(self, catalog_path: str | Path | UPath | None = None):
        """Generate parquet metadata, using the known partitions.

        If no catalog_path is provided, the catalog base directory from the `read_from_dir` call is used.

        Args:
            catalog_path (UPath): base path for the catalog

        Returns:
            sum of the number of rows in the dataset.

        Raises:
            ValueError: if no path is provided, and could not be inferred.
        """
        if catalog_path is None:
            if self.catalog_base_dir is None:
                raise ValueError("catalog_path is required if partition info was not loaded from a directory")
            catalog_path = self.catalog_base_dir

        batches = [
            [
                pa.RecordBatch.from_arrays(
                    [[pixel.order], [pixel.pixel]],
                    names=[
                        self.METADATA_ORDER_COLUMN_NAME,
                        self.METADATA_PIXEL_COLUMN_NAME,
                    ],
                )
            ]
            for pixel in self.get_healpix_pixels()
        ]

        return write_parquet_metadata_for_batches(batches, catalog_path)

    @classmethod
    def read_from_dir(cls, catalog_base_dir: str | Path | UPath | None) -> PartitionInfo:
        """Read partition info from a file within a hats directory.

        This will look for a `partition_info.csv` file, and if not found, will look for
        a `_metadata` file. The second approach is typically slower for large catalogs
        therefore a warning is issued to the user. In internal testing with large catalogs,
        the first approach takes less than a second, while the second can take 10-20 seconds.

        Args:
            catalog_base_dir: path to the root directory of the catalog

        Returns:
            A `PartitionInfo` object with the data from the file

        Raises:
            FileNotFoundError: if neither desired file is found in the catalog_base_dir
        """
        metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        if file_io.does_file_or_directory_exist(partition_info_file):
            pixel_list = PartitionInfo._read_from_csv(partition_info_file)
        elif file_io.does_file_or_directory_exist(metadata_file):
            warnings.warn("Reading partitions from parquet metadata. This is typically slow.")
            pixel_list = PartitionInfo._read_from_metadata_file(metadata_file)
        else:
            raise FileNotFoundError(
                f"_metadata or partition info file is required in catalog directory {catalog_base_dir}"
            )
        return cls(pixel_list, catalog_base_dir)

    @classmethod
    def read_from_file(cls, metadata_file: str | Path | UPath, strict: bool = False) -> PartitionInfo:
        """Read partition info from a `_metadata` file to create an object

        Args:
            metadata_file (UPath): path to the `_metadata` file
            strict (bool): use strict parsing of _metadata file. this is slower, but
                gives more helpful error messages in the case of invalid data.

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        return cls(cls._read_from_metadata_file(metadata_file, strict))

    @classmethod
    def _read_from_metadata_file(
        cls, metadata_file: str | Path | UPath, strict: bool = False
    ) -> List[HealpixPixel]:
        """Read partition info list from a `_metadata` file.

        Args:
            metadata_file (UPath): path to the `_metadata` file
            strict (bool): use strict parsing of _metadata file. this is slower, but
                gives more helpful error messages in the case of invalid data.

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        if strict:
            pixel_list = [
                HealpixPixel(
                    row_group_stat_single_value(row_group, cls.METADATA_ORDER_COLUMN_NAME),
                    row_group_stat_single_value(row_group, cls.METADATA_PIXEL_COLUMN_NAME),
                )
                for row_group in read_row_group_fragments(metadata_file)
            ]
        else:
            total_metadata = file_io.read_parquet_metadata(metadata_file)
            num_row_groups = total_metadata.num_row_groups

            first_row_group = total_metadata.row_group(0)
            norder_column = -1
            npix_column = -1

            for i in range(0, first_row_group.num_columns):
                column = first_row_group.column(i)
                if column.path_in_schema == cls.METADATA_ORDER_COLUMN_NAME:
                    norder_column = i
                elif column.path_in_schema == cls.METADATA_PIXEL_COLUMN_NAME:
                    npix_column = i

            if norder_column == -1 or npix_column == -1:
                raise ValueError("Metadata missing Norder or Npix column")

            row_group_index = np.arange(0, num_row_groups)

            pixel_list = [
                HealpixPixel(
                    total_metadata.row_group(index).column(norder_column).statistics.min,
                    total_metadata.row_group(index).column(npix_column).statistics.min,
                )
                for index in row_group_index
            ]
        ## Remove duplicates, preserving order.
        ## In the case of association partition join info, we may have multiple entries
        ## for the primary order/pixels.
        return list(dict.fromkeys(pixel_list))

    @classmethod
    def read_from_csv(cls, partition_info_file: str | Path | UPath) -> PartitionInfo:
        """Read partition info from a `partition_info.csv` file to create an object

        Args:
            partition_info_file (UPath): path to the `partition_info.csv` file

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        return cls(cls._read_from_csv(partition_info_file))

    @classmethod
    def _read_from_csv(cls, partition_info_file: str | Path | UPath) -> PartitionInfo:
        """Read partition info from a `partition_info.csv` file to create an object

        Args:
            partition_info_file (UPath): path to the `partition_info.csv` file

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(partition_info_file):
            raise FileNotFoundError(f"No partition info found where expected: {str(partition_info_file)}")

        data_frame = file_io.load_csv_to_pandas(partition_info_file)

        return [
            HealpixPixel(order, pixel)
            for order, pixel in zip(
                data_frame[cls.METADATA_ORDER_COLUMN_NAME],
                data_frame[cls.METADATA_PIXEL_COLUMN_NAME],
            )
        ]

    def as_dataframe(self):
        """Construct a pandas dataframe for the partition info pixels.

        Returns:
            Dataframe with order, directory, and pixel info.
        """
        partition_info_dict = {
            PartitionInfo.METADATA_ORDER_COLUMN_NAME: [],
            PartitionInfo.METADATA_PIXEL_COLUMN_NAME: [],
        }
        for pixel in self.pixel_list:
            partition_info_dict[PartitionInfo.METADATA_ORDER_COLUMN_NAME].append(pixel.order)
            partition_info_dict[PartitionInfo.METADATA_PIXEL_COLUMN_NAME].append(pixel.pixel)
        return pd.DataFrame.from_dict(partition_info_dict)

    @classmethod
    def from_healpix(cls, healpix_pixels: List[HealpixPixel]) -> PartitionInfo:
        """Create a partition info object from a list of constituent healpix pixels.

        Args:
            healpix_pixels: list of healpix pixels
        Returns:
            A `PartitionInfo` object with the same healpix pixels
        """
        return cls(healpix_pixels)

    def calculate_fractional_coverage(self):
        """Calculate what fraction of the sky is covered by partition tiles."""
        pixel_orders = [p.order for p in self.pixel_list]
        cov_order, cov_count = np.unique(pixel_orders, return_counts=True)
        area_by_order = [hp.nside2pixarea(hp.order2nside(order), degrees=True) for order in cov_order]
        # 41253 is the number of square degrees in a sphere
        # https://en.wikipedia.org/wiki/Square_degree
        return (area_by_order * cov_count).sum() / 41253



import pytest

from hats.loaders import read_hats


def test_read_hats_wrong_catalog_type(small_sky_dir):
    with pytest.raises(NotImplementedError, match="load catalog of type"):
        read_hats(small_sky_dir, catalog_type="unknown")


def test_read_hats_branches(
    small_sky_dir,
    small_sky_order1_dir,
    association_catalog_path,
    small_sky_source_object_index_dir,
    margin_catalog_path,
    small_sky_source_dir,
):
    read_hats(small_sky_dir)
    read_hats(small_sky_order1_dir)
    read_hats(association_catalog_path)
    read_hats(small_sky_source_object_index_dir)
    read_hats(margin_catalog_path)
    read_hats(small_sky_source_dir)



from __future__ import annotations

from mocpy import MOC
from typing_extensions import Self

import hats.pixel_math.healpix_shim as hp
from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset
from hats.pixel_tree.moc_utils import copy_moc


class MarginCatalog(HealpixDataset):
    """A HATS Catalog used to contain the 'margin' of another HATS catalog.

    Catalogs of this type are used alongside a primary catalog, and contains the margin points for each
    HEALPix pixel - any points that are within a certain distance from the HEALPix pixel boundary. This is
    used to ensure spatial operations such as crossmatching can be performed efficiently while maintaining
    accuracy.
    """

    def filter_by_moc(self, moc: MOC) -> Self:
        """Filter the pixels in the margin catalog to only include the margin pixels that overlap with the moc

        For the case of margin pixels, this includes any pixels whose margin areas may overlap with the moc.
        This is not always done with a high accuracy, but always includes any pixels that will overlap,
        and may include extra partitions that do not.

        Args:
            moc (mocpy.MOC): the moc to filter by

        Returns:
            A new margin catalog with only the pixels that overlap or that have margin area that overlap with
            the moc. Note that we reset the total_rows to None, as updating would require a scan over the new
            pixel sizes.
        """
        max_order = moc.max_order
        max_order_size = hp.nside2resol(2**max_order, arcmin=True)
        if self.catalog_info.margin_threshold > max_order_size * 60:
            raise ValueError(
                f"Cannot Filter Margin: Margin size {self.catalog_info.margin_threshold} is "
                f"greater than the size of a pixel at the highest order {max_order}."
            )
        search_moc = copy_moc(moc).add_neighbours()
        return super().filter_by_moc(search_moc)



import numpy as np

from hats.pixel_math import HealpixPixel
from hats.pixel_math.healpix_pixel import get_higher_order_pixels
from hats.pixel_tree.pixel_tree import PixelTree


def test_pixel_tree_length():
    lengths = [1, 3, 6]
    orders = [1, 2, 7]
    for order in orders:
        for length in lengths:
            hp_pixels = [(order, i) for i in range(length)]
            tree = PixelTree.from_healpix(hp_pixels)
            assert len(tree) == length


def test_get_healpix_pixels(pixel_tree_2, pixel_list_breadth_first):
    assert pixel_tree_2.get_healpix_pixels() == pixel_list_breadth_first


def test_pixel_tree_max_depth(pixel_tree_1, pixel_tree_2, pixel_tree_3):
    assert pixel_tree_1.get_max_depth() == 0
    assert pixel_tree_2.get_max_depth() == 2
    assert pixel_tree_3.get_max_depth() == 1


def test_pixel_tree_different_tree_order(pixel_tree_2):
    pixel_tree_2_order_5 = PixelTree.from_healpix(pixel_tree_2.get_healpix_pixels(), tree_order=5)
    assert pixel_tree_2_order_5.get_max_depth() == 2
    assert pixel_tree_2_order_5.tree_order == 5
    assert pixel_tree_2_order_5.get_healpix_pixels() == pixel_tree_2.get_healpix_pixels()


def test_pixel_tree_to_moc(pixel_tree_2):
    moc = pixel_tree_2.to_moc()
    moc_order_pixels = np.concatenate(
        [
            get_higher_order_pixels(pixel.order, pixel.pixel, moc.max_order - pixel.order)
            for pixel in pixel_tree_2.get_healpix_pixels()
        ]
    )
    assert np.all(moc.flatten() == moc_order_pixels)


def test_pixel_tree_contains():
    tree = PixelTree.from_healpix([(0, 0)])
    assert tree.contains((0, 0))
    assert tree.contains(HealpixPixel(0, 0))
    assert (0, 0) in tree
    assert HealpixPixel(0, 0) in tree
    assert not tree.contains((1, 1))
    assert not tree.contains(HealpixPixel(1, 1))
    assert (1, 1) not in tree
    assert HealpixPixel(1, 1) not in tree
    assert (0, 10) not in tree



"""Test pixel path creation"""

import pytest

from hats.io import paths
from hats.pixel_math.healpix_pixel import INVALID_PIXEL, HealpixPixel


def test_pixel_directory():
    """Simple case with sensical inputs"""
    expected = "/foo/dataset/Norder=0/Dir=0"
    result = paths.pixel_directory("/foo", 0, 5)
    assert str(result) == expected


def test_pixel_directory_number():
    """Simple case with sensical inputs"""
    expected = "/foo/dataset/Norder=0/Dir=0"
    result = paths.pixel_directory("/foo", pixel_order=0, pixel_number=5, directory_number=0)
    assert str(result) == expected

    result = paths.pixel_directory("/foo", pixel_order=0, directory_number=0)
    assert str(result) == expected


def test_pixel_directory_missing():
    """Simple case with missing inputs"""
    with pytest.raises(ValueError):
        paths.pixel_directory("/foo", 0)


def test_pixel_directory_nonint():
    """Simple case with non-integer inputs"""
    with pytest.raises(ValueError):
        paths.pixel_directory("/foo", "zero", "five")


def test_pixel_catalog_file():
    """Simple case with sensical inputs"""
    expected = "/foo/dataset/Norder=0/Dir=0/Npix=5.parquet"
    result = paths.pixel_catalog_file("/foo", HealpixPixel(0, 5))
    assert str(result) == expected


def test_pixel_catalog_file_w_query_params():
    expected = (
        "https://foo/dataset/Norder=0/Dir=0/Npix=5.parquet?columns=ID%2CRA%2CDEC%2Cr_auto&filters=r_auto%3C13"
    )
    query_params = {"columns": ["ID", "RA", "DEC", "r_auto"], "filters": ["r_auto<13"]}
    result = paths.pixel_catalog_file("https://foo", HealpixPixel(0, 5), query_params=query_params)
    assert expected == str(result)


def test_pixel_catalog_file_nonint():
    """Simple case with non-integer inputs"""
    with pytest.raises(AttributeError):
        paths.pixel_catalog_file("/foo", "zero", "five")


def test_pixel_catalog_files():
    expected = ["/foo/dataset/Norder=0/Dir=0/Npix=5.parquet", "/foo/dataset/Norder=1/Dir=0/Npix=16.parquet"]
    result = paths.pixel_catalog_files("/foo/", [HealpixPixel(0, 5), HealpixPixel(1, 16)])
    assert expected == result


def test_pixel_catalog_files_w_query_params():
    expected = [
        "https://foo/dataset/Norder=0/Dir=0/Npix=5.parquet?columns=ID%2CRA%2CDEC%2Cr_auto&filters=r_auto%3C13"
    ]
    query_params = {"columns": ["ID", "RA", "DEC", "r_auto"], "filters": ["r_auto<13"]}
    result = paths.pixel_catalog_files("https://foo", [HealpixPixel(0, 5)], query_params=query_params)
    assert expected == result


def test_dict_to_query_urlparams():
    expected = "?columns=ID%2CRA%2CDEC%2Cr_auto&filters=r_auto%3C13"
    query_params = {"columns": ["ID", "RA", "DEC", "r_auto"], "filters": ["r_auto<13"]}
    result = paths.dict_to_query_urlparams(query_params)
    assert result == expected

    expected = "?columns=ID%2CRA%2CDEC%2Cr_auto&filters=r_auto%3C13"
    query_params = {"columns": [" ID", "RA ", "DEC ", "r_auto"], "filters": ["r_auto < 13"]}
    result = paths.dict_to_query_urlparams(query_params)
    assert result == expected

    result = paths.dict_to_query_urlparams({})
    assert result == ""

    result = paths.dict_to_query_urlparams(None)
    assert result == ""

    result = paths.dict_to_query_urlparams({"": ""})
    assert result == ""

    result = paths.dict_to_query_urlparams({None: ""})
    assert result == ""

    result = paths.dict_to_query_urlparams({"": "nonempty"})
    assert result == ""


def test_get_healpix_from_path():
    expected = HealpixPixel(5, 34)

    # Test a few ways we could represent the path.
    result = paths.get_healpix_from_path("/foo/dataset/Norder=5/Dir=0/Npix=34.parquet")
    assert result == expected

    result = paths.get_healpix_from_path("Norder=5/Dir=0/Npix=34.pq")
    assert result == expected


def test_get_healpix_from_path_invalid():
    expected = INVALID_PIXEL

    # Test a few ways we could represent the path.
    result = paths.get_healpix_from_path("")
    assert result == expected

    result = paths.get_healpix_from_path("NORDER=5/Dir=0/PIXEL=34.pq")
    assert result == expected



import numpy as np
import pandas as pd
import pytest

from hats.io.file_io import (
    delete_file,
    load_csv_to_pandas,
    load_csv_to_pandas_generator,
    make_directory,
    read_parquet_dataset,
    read_parquet_file_to_pandas,
    remove_directory,
    write_dataframe_to_csv,
    write_string_to_file,
)
from hats.io.file_io.file_pointer import does_file_or_directory_exist


def test_make_directory(tmp_path):
    test_dir_path = tmp_path / "test_path"
    assert not does_file_or_directory_exist(test_dir_path)
    make_directory(test_dir_path)
    assert does_file_or_directory_exist(test_dir_path)


def test_make_existing_directory_raises(tmp_path):
    test_dir_path = tmp_path / "test_path"
    make_directory(test_dir_path)
    assert does_file_or_directory_exist(test_dir_path)
    with pytest.raises(OSError):
        make_directory(test_dir_path)


def test_make_existing_directory_existok(tmp_path):
    test_dir_path = tmp_path / "test_path"
    make_directory(test_dir_path)
    test_inner_dir_path = test_dir_path / "test_inner"
    make_directory(test_inner_dir_path)
    assert does_file_or_directory_exist(test_dir_path)
    assert does_file_or_directory_exist(test_inner_dir_path)
    make_directory(test_dir_path, exist_ok=True)
    assert does_file_or_directory_exist(test_dir_path)
    assert does_file_or_directory_exist(test_inner_dir_path)


def test_make_and_remove_directory(tmp_path):
    test_dir_path = tmp_path / "test_path"
    assert not does_file_or_directory_exist(test_dir_path)
    make_directory(test_dir_path)
    make_directory(test_dir_path / "subdirectory")
    (test_dir_path / "subdirectory" / "file").touch()
    assert does_file_or_directory_exist(test_dir_path)
    remove_directory(test_dir_path)
    assert not does_file_or_directory_exist(test_dir_path)

    ## Directory no longer exists to be deleted.
    with pytest.raises(FileNotFoundError):
        remove_directory(test_dir_path)

    ## Directory doesn't exist, but shouldn't throw an error.
    remove_directory(test_dir_path, ignore_errors=True)
    assert not does_file_or_directory_exist(test_dir_path)


def test_write_string_to_file(tmp_path):
    test_file_path = tmp_path / "text_file.txt"
    test_string = "this is a test"
    write_string_to_file(test_file_path, test_string, encoding="utf-8")
    with open(test_file_path, "r", encoding="utf-8") as file:
        data = file.read()
        assert data == test_string
    delete_file(test_file_path)
    assert not does_file_or_directory_exist(test_file_path)


def test_load_csv_to_pandas(small_sky_source_dir):
    partition_info_path = small_sky_source_dir / "partition_info.csv"
    frame = load_csv_to_pandas(partition_info_path)
    assert len(frame) == 14


def test_load_csv_to_pandas_generator(small_sky_source_dir):
    partition_info_path = small_sky_source_dir / "partition_info.csv"
    num_reads = 0
    for frame in load_csv_to_pandas_generator(partition_info_path, chunksize=7, compression=None):
        assert len(frame) == 7
        num_reads += 1
    assert num_reads == 2


def test_write_df_to_csv(tmp_path):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = tmp_path / "test.csv"
    write_dataframe_to_csv(random_df, test_file_path, index=False)
    loaded_df = pd.read_csv(test_file_path)
    pd.testing.assert_frame_equal(loaded_df, random_df)


def test_read_parquet_data(tmp_path):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = tmp_path / "test.parquet"
    random_df.to_parquet(test_file_path)
    dataframe = read_parquet_file_to_pandas(test_file_path)
    pd.testing.assert_frame_equal(dataframe, random_df)


def test_read_parquet_dataset(small_sky_dir, small_sky_order1_dir):
    (_, ds) = read_parquet_dataset(small_sky_dir / "dataset" / "Norder=0")

    assert ds.count_rows() == 131

    (_, ds) = read_parquet_dataset([small_sky_dir / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"])

    assert ds.count_rows() == 131

    (_, ds) = read_parquet_dataset(
        [
            small_sky_order1_dir / "dataset" / "Norder=1" / "Dir=0" / "Npix=44.parquet",
            small_sky_order1_dir / "dataset" / "Norder=1" / "Dir=0" / "Npix=45.parquet",
            small_sky_order1_dir / "dataset" / "Norder=1" / "Dir=0" / "Npix=46.parquet",
            small_sky_order1_dir / "dataset" / "Norder=1" / "Dir=0" / "Npix=47.parquet",
        ]
    )

    assert ds.count_rows() == 131



from __future__ import annotations

import warnings
from pathlib import Path

import numpy as np
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.partition_info import PartitionInfo
from hats.io import get_common_metadata_pointer, get_parquet_metadata_pointer, get_partition_info_pointer
from hats.io.file_io import read_parquet_dataset
from hats.io.file_io.file_pointer import is_regular_file
from hats.io.paths import get_healpix_from_path
from hats.loaders import read_hats
from hats.pixel_math.healpix_pixel import INVALID_PIXEL
from hats.pixel_math.healpix_pixel_function import sort_pixels


# pylint: disable=too-many-statements,too-many-locals
def is_valid_catalog(
    pointer: str | Path | UPath,
    strict: bool = False,
    fail_fast: bool = False,
    verbose: bool = True,
) -> bool:
    """Checks if a catalog is valid for a given base catalog pointer

    Args:
        pointer (UPath): pointer to base catalog directory
        strict (bool): should we perform additional checking that every optional
            file exists, and contains valid, consistent information.
        fail_fast (bool): if performing strict checks, should we return at the first
            failure, or continue and find all problems?
        verbose (bool): if performing strict checks, should we print out counts,
            progress, and approximate sky coverage?

    Returns:
        True if both the properties and partition_info files are
        valid, False otherwise
    """
    if not strict:
        return is_catalog_info_valid(pointer) and (
            is_partition_info_valid(pointer) or is_metadata_valid(pointer)
        )

    if verbose:
        print(f"Validating catalog at path {pointer} ... ")

    is_valid = True

    def handle_error(msg):
        """inline-method to handle repeated logic of raising error or warning and
        continuing."""
        nonlocal fail_fast
        nonlocal is_valid
        nonlocal verbose
        if fail_fast:
            raise ValueError(msg)
        if verbose:
            print(msg)
        else:
            warnings.warn(msg)
        is_valid = False

    if not is_catalog_info_valid(pointer):
        handle_error("properties file does not exist or is invalid.")

    if not is_partition_info_valid(pointer):
        handle_error("partition_info.csv file does not exist.")

    if not is_metadata_valid(pointer):
        handle_error("_metadata file does not exist.")

    if not is_common_metadata_valid(pointer):
        handle_error("_common_metadata file does not exist.")

    if not is_valid:
        # Even if we're not failing fast, we need to stop here if the metadata
        # files don't exist.
        return is_valid

    # Load as a catalog object. Confirms that the catalog info matches type.
    catalog = read_hats(pointer)
    expected_pixels = sort_pixels(catalog.get_healpix_pixels())

    if verbose:
        print(f"Found {len(expected_pixels)} partitions.")

    ## Compare the pixels in _metadata with partition_info.csv
    metadata_file = get_parquet_metadata_pointer(pointer)

    # Use both strategies of reading the partition info: strict and !strict.
    metadata_pixels = sort_pixels(
        PartitionInfo.read_from_file(metadata_file, strict=True).get_healpix_pixels()
    )
    if not np.array_equal(expected_pixels, metadata_pixels):
        handle_error("Partition pixels differ between catalog and _metadata file (strict)")

    metadata_pixels = sort_pixels(
        PartitionInfo.read_from_file(metadata_file, strict=False).get_healpix_pixels()
    )
    if not np.array_equal(expected_pixels, metadata_pixels):
        handle_error("Partition pixels differ between catalog and _metadata file (non-strict)")

    partition_info_file = get_partition_info_pointer(pointer)
    csv_pixels = sort_pixels(PartitionInfo.read_from_csv(partition_info_file).get_healpix_pixels())
    if not np.array_equal(expected_pixels, csv_pixels):
        handle_error("Partition pixels differ between catalog and partition_info.csv file")

    ## Load as parquet dataset. Allow errors, and check pixel set against _metadata
    ignore_prefixes = [
        "_common_metadata",
        "_metadata",
        "catalog_info.json",
        "properties",
        "provenance_info.json",
        "partition_info.csv",
        "point_map.fits",
        "README",
    ]

    # As a side effect, this confirms that we can load the directory as a valid dataset.
    (dataset_path, dataset) = read_parquet_dataset(
        pointer,
        ignore_prefixes=ignore_prefixes,
        exclude_invalid_files=False,
    )

    parquet_path_pixels = []
    for hats_file in dataset.files:
        relative_path = hats_file[len(dataset_path) :]
        healpix_pixel = get_healpix_from_path(relative_path)
        if healpix_pixel == INVALID_PIXEL:
            handle_error(f"Could not derive partition pixel from parquet path: {relative_path}")

        parquet_path_pixels.append(healpix_pixel)

    parquet_path_pixels = sort_pixels(parquet_path_pixels)

    if not np.array_equal(expected_pixels, parquet_path_pixels):
        handle_error("Partition pixels differ between catalog and parquet paths")

    if verbose:
        # Print a few more stats
        pixel_orders = [p.order for p in expected_pixels]
        cov_order, cov_count = np.unique(pixel_orders, return_counts=True)
        area_by_order = [hp.nside2pixarea(hp.order2nside(order), degrees=True) for order in cov_order]
        total_area = (area_by_order * cov_count).sum()
        print(
            f"Approximate coverage is {total_area:0.2f} sq deg, or {total_area/41253*100:0.2f} % of the sky."
        )

    return is_valid


def is_catalog_info_valid(pointer: str | Path | UPath) -> bool:
    """Checks if properties file is valid for a given base catalog pointer

    Args:
        pointer (UPath): pointer to base catalog directory

    Returns:
        True if the properties file exists, and it is correctly formatted,
        False otherwise
    """
    try:
        TableProperties.read_from_dir(pointer)
    except (FileNotFoundError, ValueError, NotImplementedError):
        return False
    return True


def is_partition_info_valid(pointer: str | Path | UPath) -> bool:
    """Checks if partition_info is valid for a given base catalog pointer

    Args:
        pointer (UPath): pointer to base catalog directory

    Returns:
        True if the partition_info file exists, False otherwise
    """
    partition_info_pointer = get_partition_info_pointer(pointer)
    partition_info_exists = is_regular_file(partition_info_pointer)
    return partition_info_exists


def is_metadata_valid(pointer: str | Path | UPath) -> bool:
    """Checks if _metadata is valid for a given base catalog pointer

    Args:
        pointer (UPath): pointer to base catalog directory

    Returns:
        True if the _metadata file exists, False otherwise
    """
    metadata_file = get_parquet_metadata_pointer(pointer)
    metadata_file_exists = is_regular_file(metadata_file)
    return metadata_file_exists


def is_common_metadata_valid(pointer: str | Path | UPath) -> bool:
    """Checks if _common_metadata is valid for a given base catalog pointer

    Args:
        pointer (UPath): pointer to base catalog directory

    Returns:
        True if the _common_metadata file exists, False otherwise
    """
    metadata_file = get_common_metadata_pointer(pointer)
    metadata_file_exists = is_regular_file(metadata_file)
    return metadata_file_exists



"""Tests of pixel margin utility functions"""

import numpy as np
import numpy.testing as npt
import pytest

import hats.pixel_math.pixel_margins as pm


def test_get_margin():
    """Check that the code works in the standard case."""
    margins = pm.get_margin(2, 2, 2)

    expected = np.array(
        [
            1141,
            1143,
            1149,
            1151,
            1237,
            128,
            129,
            132,
            133,
            144,
            48,
            50,
            56,
            58,
            26,
            10,
            11,
            14,
            15,
            1119,
        ]
    )

    assert len(margins) == 20

    npt.assert_array_equal(margins, expected)


def test_zero_dk():
    """Check that the code fails when trying to find margins at same order as the pixel."""
    with pytest.raises(ValueError) as value_error:
        pm.get_margin(2, 2, 0)

    assert str(value_error.value) == "dk must be greater than order"


def test_negative_dk():
    """Check that the code fails when trying to find margins at a higher order than the pixel."""
    with pytest.raises(ValueError) as value_error:
        pm.get_margin(2, 2, -1)

    assert str(value_error.value) == "dk must be greater than order"


def test_polar_edge():
    """Check that the code works when trying to find margins around the north pole."""
    margins = pm.get_margin(2, 5, 2)

    expected = np.array(
        [
            69,
            71,
            77,
            79,
            101,
            112,
            113,
            116,
            117,
            442,
            426,
            427,
            430,
            431,
            1530,
            1531,
            1534,
            1535,
            1519,
        ]
    )

    assert len(margins) == 19

    npt.assert_array_equal(margins, expected)


def test_polar_edge_south():
    """Check that the code works when trying to find margins around the south pole."""
    margins = pm.get_margin(1, 35, 2)

    expected = np.array(
        [
            549,
            551,
            557,
            559,
            261,
            272,
            273,
            276,
            277,
            0,
            352,
            354,
            360,
            362,
            330,
            538,
            539,
            542,
            543,
            527,
        ]
    )
    assert len(margins) == 20

    npt.assert_array_equal(margins, expected)


def test_edge_negative_value():
    """Check to make sure get_edge fails when passed a negative edge value."""
    with pytest.raises(ValueError) as value_error:
        pm.get_edge(2, 5, -1)

    assert str(value_error.value) == "edge can only be values between 0 and 7 (see docstring)"


def test_edge_greater_than_7():
    """Check to make sure get_edge fails when passed an edge value greater than 7."""
    with pytest.raises(ValueError) as value_error:
        pm.get_edge(2, 5, 8)

    assert str(value_error.value) == "edge can only be values between 0 and 7 (see docstring)"
